# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import shutil
import multiprocessing
import datetime
from collections import defaultdict

import six

from . import Command
from ..benchmarks import Benchmarks
from ..console import log
from ..graph import GraphSet
from ..machine import iter_machine_files
from ..repo import get_repo
from ..results import iter_results
from ..publishing import OutputPublisher
from .. import statistics
from .. import util
from .. import __version__


def check_benchmark_params(name, benchmark):
    """
    Check benchmark params and param_keys items, so that the javascript can
    assume this data is valid. It is checked in benchmark.py already when it
    is generated, but best to double check in any case.
    """
    if 'params' not in benchmark:
        # Old-format benchmarks.json
        benchmark['params'] = []
        benchmark['param_names'] = []

    msg = "Information in benchmarks.json for benchmark %s is malformed" % (
        name)
    if (not isinstance(benchmark['params'], list) or
        not isinstance(benchmark['param_names'], list)):
        raise ValueError(msg)
    if len(benchmark['params']) != len(benchmark['param_names']):
        raise ValueError(msg)
    for item in benchmark['params']:
        if not isinstance(item, list):
            raise ValueError(msg)


class Publish(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "publish", help="Collate results into a website",
            description=
            """
            Collate all results into a website.  This website will be
            written to the ``html_dir`` given in the ``asv.conf.json``
            file, and may be served using any static web server.""")
        parser.add_argument(
            '--no-pull', action='store_true', dest='no_pull',
            help="Do not pull the repository")
        parser.add_argument(
            'range', nargs='?', default=None,
            help="""Optional commit range to consider""")
        parser.add_argument(
            '--html-dir', '-o', default=None, help=(
                "Optional output directory. Default is 'html_dir' "
                "from asv config"))

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args):
        if args.html_dir is not None:
            conf.html_dir = args.html_dir
        return cls.run(conf=conf, range_spec=args.range, pull=not args.no_pull)

    @staticmethod
    def iter_results(conf, repo, range_spec=None):
        if range_spec is not None:
            if isinstance(range_spec, list):
                hashes = range_spec
            else:
                hashes = repo.get_hashes_from_range(range_spec)
        else:
            hashes = None
        for result in iter_results(conf.results_dir):
            if hashes is None or result.commit_hash in hashes:
                yield result

    @classmethod
    def run(cls, conf, range_spec=None, pull=True):
        params = {}
        env_vars = defaultdict(set)
        graphs = GraphSet()
        machines = {}
        benchmark_names = set()

        log.set_nitems(6 + len(list(util.iter_subclasses(OutputPublisher))))

        if os.path.exists(conf.html_dir):
            util.long_path_rmtree(conf.html_dir)

        repo = get_repo(conf)
        benchmarks = Benchmarks.load(conf)

        def copy_ignore(src, names):
            # Copy only *.js and *.css in vendor dir
            ignore = [fn for fn in names
                      if (os.path.basename(src).lower() == 'vendor' and
                          not fn.lower().endswith('.js') and
                          not fn.lower().endswith('.css'))]
            return ignore

        template_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), '..', 'www')
        shutil.copytree(template_dir, conf.html_dir, ignore=copy_ignore)

        # Ensure html_dir is writable even if template_dir is on a read-only FS
        os.chmod(conf.html_dir, 0o755)
        for (pre, ds, fs) in os.walk(conf.html_dir):
            for x in fs:
                os.chmod(os.path.join(pre, x), 0o644)
            for x in ds:
                os.chmod(os.path.join(pre, x), 0o755)

        log.step()
        log.info("Loading machine info")
        with log.indent():
            for path in iter_machine_files(conf.results_dir):
                d = util.load_json(path)
                machines[d['machine']] = d

        log.step()
        log.info("Getting params, commits, tags and branches")
        with log.indent():
            # Determine first the set of all parameters and all commits
            hash_to_date = {}
            for results in cls.iter_results(conf, repo, range_spec):
                hash_to_date[results.commit_hash] = results.date
                for key, val in six.iteritems(results.params):
                    if val is None:
                        # Backward compatibility -- null means ''
                        val = ''

                    params.setdefault(key, set())
                    params[key].add(val)

                for name, val in six.iteritems(results.env_vars):
                    # Prefix them in case of name collision
                    env_vars["env-{}".format(name)].add(val)

            params.update(env_vars)

            if pull:
                repo.pull()
            tags = repo.get_tags()
            revisions = repo.get_revisions(set(hash_to_date.keys()) | set(tags.values()))

            for tag, commit_hash in list(tags.items()):
                # Map to revision number instead of commit hash and add tags to hash_to_date
                tags[tag] = revisions[tags[tag]]
                hash_to_date[commit_hash] = repo.get_date_from_name(commit_hash)

            revision_to_date = dict((r, hash_to_date[h]) for h, r in six.iteritems(revisions))

            branches = dict(
                (branch, repo.get_branch_commits(branch))
                for branch in conf.branches)

        log.step()
        log.info("Loading results")
        with log.indent():
            # Generate all graphs
            for results in cls.iter_results(conf, repo, range_spec):
                log.dot()

                branches_for_commit = [branch for branch, commits in branches.items() if
                                       results.commit_hash in commits]

                # Print a warning message if we couldn't find the branch of a commit
                if not len(branches_for_commit):
                    msg = "Couldn't find {} in branches ({})"
                    log.warning(msg.format(results.commit_hash[:conf.hash_length],
                                           ", ".join(str(branch) for branch in branches.keys())))

                for key in results.get_result_keys(benchmarks):
                    b = benchmarks[key]
                    b_params = b['params']

                    result = results.get_result_value(key, b_params)
                    weight = [statistics.get_weight(s)
                              for s in results.get_result_stats(key, b_params)]
                    if not b_params:
                        result = result[0]
                        weight = weight[0]

                    benchmark_names.add(key)

                    for branch in branches_for_commit:
                        cur_params = dict(results.params)
                        cur_env = {'env-{}'.format(name): val
                                   for name, val in six.iteritems(results.env_vars)}
                        cur_params.update(cur_env)
                        cur_params['branch'] = repo.get_branch_name(branch)

                        # Backward compatibility, see above
                        for param_key, param_value in list(cur_params.items()):
                            if param_value is None:
                                cur_params[param_key] = ''

                        # Fill in missing params
                        for param_key in params.keys():
                            if param_key not in cur_params:
                                cur_params[param_key] = None
                                params[param_key].add(None)

                        # Create graph
                        graph = graphs.get_graph(key, cur_params)
                        graph.add_data_point(revisions[results.commit_hash], result, weight)

            # Get the parameter sets for all graphs
            graph_param_list = []
            for path, graph in graphs:
                if 'summary' not in graph.params:
                    if graph.params not in graph_param_list:
                        graph_param_list.append(graph.params)

        log.step()
        log.info("Detecting steps")
        with log.indent():
            n_processes = multiprocessing.cpu_count()
            pool = util.get_multiprocessing_pool(n_processes)
            try:
                graphs.detect_steps(pool, dots=log.dot)
                pool.close()
                pool.join()
            finally:
                pool.terminate()

        log.step()
        log.info("Generating graphs")
        with log.indent():
            # Save files
            graphs.save(conf.html_dir, dots=log.dot)

        pages = []
        classes = sorted(util.iter_subclasses(OutputPublisher),
                         key=lambda cls: cls.order)
        for cls in classes:
            log.step()
            log.info("Generating output for {0}".format(cls.__name__))
            with log.indent():
                cls.publish(conf, repo, benchmarks, graphs, revisions)
                pages.append([cls.name, cls.button_label, cls.description])

        log.step()
        log.info("Writing index")
        benchmark_map = dict(benchmarks)
        for key in six.iterkeys(benchmark_map):
            check_benchmark_params(key, benchmark_map[key])
        for key, val in six.iteritems(params):
            val = list(val)
            val.sort(key=lambda x: '[none]' if x is None else str(x))
            params[key] = val
        params['branch'] = [repo.get_branch_name(branch) for branch in conf.branches]
        revision_to_hash = dict((r, h) for h, r in six.iteritems(revisions))
        util.write_json(os.path.join(conf.html_dir, "index.json"), {
            'project': conf.project,
            'project_url': conf.project_url,
            'show_commit_url': conf.show_commit_url,
            'hash_length': conf.hash_length,
            'revision_to_hash': revision_to_hash,
            'revision_to_date': revision_to_date,
            'params': params,
            'graph_param_list': graph_param_list,
            'benchmarks': benchmark_map,
            'machines': machines,
            'tags': tags,
            'pages': pages,
        }, compact=True)

        util.write_json(os.path.join(conf.html_dir, "info.json"), {
            'asv-version': __version__,
            'timestamp': util.datetime_to_js_timestamp(datetime.datetime.utcnow())
        })



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import sys

# This __version__ assignment is parsed by setup.py; keep it in this form.
# Development versions end with ".dev" (suffix is added below).
__version__ = "0.5.1"
__release__ = not __version__.endswith(".dev")

try:
    from ._version import __githash__, __suffix__
except ImportError:
    __githash__ = None
    __suffix__ = "0"
if not __release__:
    __version__ += __suffix__
del __suffix__


if sys.version_info >= (3, 3):
    # OS X framework builds of Python 3.3 can not call other 3.3
    # virtualenvs as a subprocess because `__PYENV_LAUNCHER__` is
    # inherited.
    if os.environ.get('__PYVENV_LAUNCHER__'):
        os.unsetenv('__PYVENV_LAUNCHER__')

from . import plugin_manager



'''
Created on 20/01/2011

v0.2 (C) Gerald Storer
MIT License

Based on JSON.minify.js:
https://github.com/getify/JSON.minify

Contributers:
 - Pradyun S. Gedam (conditions and variable names changed)
'''


import re

def json_minify(string, strip_space=True):
    tokenizer = re.compile(r'"|(/\*)|(\*/)|(//)|\n|\r')
    in_string = False
    in_multi = False
    in_single = False

    new_str = []
    index = 0

    for match in re.finditer(tokenizer, string):

        if not (in_multi or in_single):
            tmp = string[index:match.start()]
            if not in_string and strip_space:
                # replace white space as defined in standard
                tmp = re.sub('[ \t\n\r]+', '', tmp)
            new_str.append(tmp)

        index = match.end()
        val = match.group()

        if val == '"' and not (in_multi or in_single):
            escaped = re.search(r'(\\)*$', string[:match.start()])

            # start of string or unescaped quote character to end string
            if not in_string or (escaped is None or len(escaped.group()) % 2 == 0):
                in_string = not in_string
            index -= 1 # include " character in next catch
        elif not (in_string or in_multi or in_single):
            if val == '/*':
                in_multi = True
            elif val == '//':
                in_single = True
        elif val == '*/' and in_multi and not (in_string or in_single):
            in_multi = False
        elif val in '\r\n' and not (in_multi or in_string) and in_single:
            in_single = False
        elif not ((in_multi or in_single) or (val in ' \r\n\t' and strip_space)):
            new_str.append(val)

    new_str.append(string[index:])
    return ''.join(new_str)


if __name__ == '__main__':
    # Python 2.6+ needed to run tests
    import json
    import textwrap
    import unittest

    class JsonMinifyTestCase(unittest.TestCase):
        """Tests for json_minify"""
        def template(self, in_string, expected):
            in_dict = json.loads(json_minify(in_string))
            expected_dict = json.loads(expected)
            self.assertEqual(in_dict, expected_dict)

        def test_1(self):
            self.template(textwrap.dedent('''
                // this is a JSON file with comments
                {
                    "foo": "bar",    // this is cool
                    "bar": [
                        "baz", "bum"
                    ],
                /* the rest of this document is just fluff
                   in case you are interested. */
                    "something": 10,
                    "else": 20
                }

                /* NOTE: You can easily strip the whitespace and comments
                   from such a file with the JSON.minify() project hosted
                   here on github at http://github.com/getify/JSON.minify
                */'''),
                '{"foo":"bar","bar":["baz","bum"],"something":10,"else":20}'
            )

        def test_2(self):
            self.template(textwrap.dedent('''
                {"/*":"*/","//":"",/*"//"*/"/*/"://
                "//"}'''),
                '{"/*":"*/","//":"","/*/":"//"}'
            )

        def test_3(self):
            self.template(textwrap.dedent(r'''
                /*
                this is a
                multi line comment */{

                "foo"
                :
                    "bar/*"// something
                    ,    "b\"az":/*
                something else */"blah"

                }
                '''),
                r'{"foo":"bar/*","b\"az":"blah"}'
            )

        def test_4(self):
            self.template(textwrap.dedent(r'''
                {"foo": "ba\"r//", "bar\\": "b\\\"a/*z",
                "baz\\\\": /* yay */ "fo\\\\\"*/o"
                }
                '''),
                r'{"foo":"ba\"r//","bar\\":"b\\\"a/*z","baz\\\\":"fo\\\\\"*/o"}'
            )

    unittest.main()



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import contextlib
import io
import os
import pstats
import sys
import tempfile

from . import Command
from ..benchmarks import Benchmarks
from ..console import log, color_print
from ..environment import get_environments, is_existing_only
from ..machine import Machine
from ..profiling import ProfilerGui
from ..repo import get_repo, NoSuchNameError
from ..results import iter_results_for_machine
from ..runner import run_benchmarks
from ..util import hash_equal, iter_subclasses
from .. import util

from . import common_args


@contextlib.contextmanager
def temp_profile(profile_data):
    profile_fd, profile_path = tempfile.mkstemp()
    try:
        with io.open(profile_fd, 'wb', closefd=True) as fd:
            fd.write(profile_data)

        yield profile_path
    finally:
        os.remove(profile_path)


class Profile(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "profile",
            help="""Run the profiler on a particular benchmark on a
            particular revision""",
            description="Profile a benchmark")

        parser.add_argument(
            'benchmark',
            help="""The benchmark to profile.  Must be a
            fully-specified benchmark name. For parameterized benchmark, it
            must include the parameter combination to use, e.g.:
            benchmark_name(param0, param1, ...)""")
        parser.add_argument(
            'revision', nargs='?',
            help="""The revision of the project to profile.  May be a
            commit hash, or a tag or branch name.""")
        parser.add_argument(
            '--gui', '-g',
            help="""Display the profile in the given gui.  Use
            --gui=list to list available guis.""")
        parser.add_argument(
            '--output', '-o',
            help="""Save the profiling information to the given file.
            This file is in the format written by the `cProfile`
            standard library module.  If not provided, prints a simple
            text-based profiling report to the console.""")
        parser.add_argument(
            '--force', '-f', action='store_true',
            help="""Forcibly re-run the profile, even if the data
            already exists in the results database.""")
        common_args.add_environment(parser)
        common_args.add_launch_method(parser)

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def find_guis(cls):
        cls.guis = {}
        for x in iter_subclasses(ProfilerGui):
            if x.name is not None and x.is_available():
                cls.guis[x.name] = x

    @classmethod
    def run_from_conf_args(cls, conf, args, **kwargs):
        return cls.run(
            conf=conf, benchmark=args.benchmark, revision=args.revision,
            gui=args.gui, output=args.output, force=args.force,
            env_spec=args.env_spec, launch_method=args.launch_method,
            **kwargs)

    @classmethod
    def run(cls, conf, benchmark, revision=None, gui=None, output=None,
            force=False, env_spec=None, launch_method=None,
            _machine_file=None):
        cls.find_guis()

        if gui == 'list':
            log.info("Available profiler GUIs:")
            with log.indent():
                for x in cls.guis.values():
                    log.info("{0}: {1}".format(x.name, x.description))
            return

        if gui is not None and gui not in cls.guis:
            raise util.UserError(
                "Unknown profiler GUI {0}".format(gui))

        if benchmark is None:
            raise util.UserError(
                "Must specify benchmark to run")

        environments = list(get_environments(conf, env_spec))

        if is_existing_only(environments):
            # No repository required, so skip using it
            conf.dvcs = "none"

        repo = get_repo(conf)
        repo.pull()

        machine_name = Machine.get_unique_machine_name()
        if revision is None:
            rev = conf.branches[0]
        else:
            rev = revision

        try:
            commit_hash = repo.get_hash_from_name(rev)
        except NoSuchNameError as exc:
            raise util.UserError("Unknown commit {0}".format(exc))

        profile_data = None
        checked_out = set()

        # First, we see if we already have the profile in the results
        # database
        if not force and commit_hash:
            for result in iter_results_for_machine(
                    conf.results_dir, machine_name):
                if hash_equal(commit_hash, result.commit_hash):
                    if result.has_profile(benchmark):
                        env_matched = any(result.env.name == env.name
                                          for env in environments)
                        if env_matched:
                            if result.env.name not in checked_out:
                                # We need to checkout the correct commit so that
                                # the line numbers in the profile data match up with
                                # what's in the source tree.
                                result.env.checkout_project(repo, commit_hash)
                                checked_out.add(result.env.name)
                            profile_data = result.get_profile(benchmark)
                            break

        if profile_data is None:
            if len(environments) == 0:
                log.error("No environments selected")
                return

            if revision is not None:
                for env in environments:
                    if not env.can_install_project():
                        raise util.UserError(
                            "An explicit revision may not be specified when "
                            "using an existing environment.")

            env = environments[0]

            if env.python != "{0}.{1}".format(*sys.version_info[:2]):
                raise util.UserError(
                    "Profiles must be run in the same version of Python as the "
                    "asv master process")

            benchmarks = Benchmarks.discover(conf, repo, environments,
                                             [commit_hash],
                                             regex='^{0}$'.format(benchmark))

            if len(benchmarks) == 0:
                raise util.UserError("'{0}' benchmark not found".format(benchmark))
            elif len(benchmarks) > 1:
                exact_matches = benchmarks.filter_out([x for x in benchmarks if x != benchmark])
                if len(exact_matches) == 1:
                    log.warning("'{0}' matches more than one benchmark, using exact match".format(benchmark))
                    benchmarks = exact_matches
                else:
                    raise util.UserError("'{0}' matches more than one benchmark".format(benchmark))

            benchmark_name, = benchmarks.keys()

            if not force:
                log.info(
                    "Profile data does not already exist. "
                    "Running profiler now.")
            else:
                log.info("Running profiler")
            with log.indent():
                env.install_project(conf, repo, commit_hash)

                results = run_benchmarks(
                    benchmarks, env, show_stderr=True, quick=False, profile=True,
                    launch_method=launch_method)

                profile_data = results.get_profile(benchmark_name)

        log.flush()

        if gui is not None:
            log.debug("Opening gui {0}".format(gui))
            with temp_profile(profile_data) as profile_path:
                return cls.guis[gui].open_profiler_gui(profile_path)
        elif output is not None:
            with io.open(output, 'wb') as fd:
                fd.write(profile_data)
        else:
            color_print('')
            with temp_profile(profile_data) as profile_path:
                stats = pstats.Stats(profile_path)
                stats.sort_stats('cumulative')
                stats.print_stats()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, unicode_literals, print_function

import os
import itertools
import six

from ..console import log
from ..publishing import OutputPublisher
from ..graph import Graph

from .. import util



def benchmark_param_iter(benchmark):
    """
    Iterate over all combinations of parameterized benchmark parameters.

    Yields
    ------
    idx : int
        Combination flat index. `None` if benchmark not parameterized.
    params : tuple
        Tuple of parameter values.

    """
    if not benchmark['params']:
        yield None, ()
    else:
        for item in enumerate(itertools.product(*benchmark['params'])):
            yield item


class SummaryList(OutputPublisher):
    name = "summarylist"
    button_label = "List view"
    description = "Display as a list"
    order = 1

    @classmethod
    def publish(cls, conf, repo, benchmarks, graphs, revisions):
        results = {}

        # Investigate all benchmarks
        for benchmark_name, benchmark in sorted(six.iteritems(benchmarks)):
            log.dot()

            benchmark_graphs = graphs.get_graph_group(benchmark_name)

            # For parameterized benchmarks, consider each combination separately
            for idx, benchmark_param in benchmark_param_iter(benchmark):
                pretty_name = benchmark_name

                if benchmark.get('pretty_name'):
                    pretty_name = benchmark['pretty_name']

                if idx is not None:
                    pretty_name = '{0}({1})'.format(pretty_name,
                                                    ", ".join(benchmark_param))

                # Each environment parameter combination is reported
                # separately on the summarylist page
                benchmark_graphs = graphs.get_graph_group(benchmark_name)
                for graph in benchmark_graphs:
                    # Produce interesting information, based on
                    # stepwise fit on the benchmark data (reduces noise)
                    steps = graph.get_steps()
                    if idx is not None and steps:
                        steps = graph.get_steps()[idx]

                    last_value = None
                    last_err = None
                    change_rev = None
                    last_rev = None
                    prev_value = None

                    if not steps:
                        # No data
                        pass
                    else:
                        last_piece = steps[-1]
                        last_value = last_piece[2]
                        last_err = last_piece[4]
                        last_rev = last_piece[1] - 1

                        if len(steps) > 1:
                            prev_piece = steps[-2]
                            prev_value = prev_piece[2]
                            if prev_piece[1] == last_piece[0]:
                                # Single commit
                                change_rev = [None, last_piece[0]]
                            else:
                                # Revision range (left-exclusive)
                                change_rev = [prev_piece[1]-1, last_piece[0]]

                    row = dict(name=benchmark_name,
                               idx=idx,
                               pretty_name=pretty_name,
                               last_rev=last_rev,
                               last_value=last_value,
                               last_err=last_err,
                               prev_value=prev_value,
                               change_rev=change_rev)

                    # Generate summary data filename.
                    # Note that 'summary' is not a valid benchmark name, so that we can
                    # be sure it can be always used.
                    path = Graph.get_file_path(graph.params, 'summary') + ".json"
                    results.setdefault(path, []).append(row)

        # Write results to files
        for path, data in six.iteritems(results):
            filename = os.path.join(conf.html_dir, path)
            util.write_json(filename, sorted(data, key=lambda x: (x['name'], x['idx'])),
                            compact=True)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from .run import Run

from . import common_args


class Dev(Run):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "dev", help="Do a test run of a benchmark suite during development",
            description="""
                This runs a benchmark suite in a mode that is useful
                during development.  It is equivalent to
                ``asv run --python=same``""")

        cls._setup_arguments(parser, env_default_same=True)

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run(cls, conf, **kwargs):
        if not kwargs.get("env_spec"):
            kwargs["env_spec"] = ["existing:same"]
        return super(cls, Dev).run(conf=conf, **kwargs)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import sys
import pkgutil

from . import commands
from . import plugins


class PluginManager(object):
    """
    A class to load and manage plugins.

    By default in asv, plugins are searched for in the `asv.plugins`
    namespace package and in the `asv.commands` package.

    Then, any modules specified in the ``plugins`` entry in the
    ``asv.conf.json`` file are loaded.
    """
    def __init__(self):
        self._plugins = []

    def load_plugins(self, package):
        prefix = package.__name__ + '.'
        for module_finder, name, ispkg in pkgutil.iter_modules(package.__path__, prefix):
            __import__(name)
            mod = sys.modules[name]
            self.init_plugin(mod)
            self._plugins.append(mod)

    def import_plugin(self, name):
        extended = False
        if name.startswith('.'):
            extended = True
            sys.path.insert(0, '.')
            name = name[1:]
        try:
            mod = __import__(name, {}, {}, [], level=0)
            self.init_plugin(mod)
            self._plugins.append(mod)
        finally:
            if extended:
                del sys.path[0]

    def init_plugin(self, mod):
        if hasattr(mod, 'setup'):
            mod.setup()

    def run_hook(self, hook_name, args, kwargs):
        for plugin in self._plugins:
            if hasattr(plugin, hook_name):
                getattr(plugin, hook_name)(*args, **kwargs)


plugin_manager = PluginManager()
plugin_manager.load_plugins(commands)
plugin_manager.load_plugins(plugins)

commands.__doc__ = commands._make_docstring()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
r"""Regression detection in ASV is based on detecting stepwise changes
in the graphs. The assumptions on the data are as follows: the curves
are piecewise constant plus random noise. We don't know the scaling of
the data or the amplitude of the noise, but assume the relative weight
of the noise amplitude is known for each data point.

ASV measures the noise amplitude of each data point, based on a number
of samples. We use this information for weighting the different data
points:

.. math::

   \sigma_j = \sigma \mathrm{CI}_{99} = \sigma / w_j

i.e., we assume the uncertainty in each measurement point is
proportional to the estimated confidence interval for each data point.
Their inverses are taken as the relative weights ``w_j``. If ``w_j=0``
or undefined, we replace it with the median weight, or with ``1`` if
all are undefined. The step detection algorithm determines the
absolute noise amplitude itself based on all available data, which is
more robust than relying on the individual measurements.

Step detection is a well-studied problem. In this implementation, we
mainly follow a variant of the approach outlined in [Friedrich2008]_
and elsewhere. This provides a fast algorithm for solving the
piecewise weighted fitting problem

.. math::
   :label: gamma-opt

   \mathop{\mathrm{argmin}}_{k,\{j\},\{\mu\}} \gamma k + \sum_{r=1}^k\sum_{i=j_{r-1}}^{j_r} w_i |y_i - \mu_r|

The differences are: as we do not need exact solutions, we add
additional heuristics to work around the :math:`{\mathcal O}(n^2)`
scaling, which is too harsh for pure-Python code. For details, see
``asv.step_detect.solve_potts_approx``.  Moreover, we follow a
slightly different approach on obtaining a suitable number of
intervals, by selecting an optimal value for :math:`\gamma`, based on
a variant of the information criterion problem discussed in
[Yao1988]_.


.. [Friedrich2008] F. Friedrich et al.,
   ''Complexity Penalized M-Estimation: Fast Computation'',
   Journal of Computational and Graphical Statistics 17.1, 201-224 (2008).
   http://dx.doi.org/10.1198/106186008X285591

.. [Yao1988] Y.-C. Yao,
   ''Estimating the number of change-points via Schwarz criterion'',
   Statistics & Probability Letters 6, 181-189 (1988).
   http://dx.doi.org/10.1016/0167-7152(88)90118-6


Bayesian information
--------------------

To proceed, we need an argument by which to select a suitable
:math:`\gamma` in :eq:`gamma-opt`. Some of the literature on step
detection, e.g. [Yao1988]_, suggests results based on Schwarz information
criteria,

.. math::
   :label: ic-form

   \text{SC} = \frac{m}{2} \ln \sigma^2 + k \ln m = \text{min!}

where :math:`\sigma^2` is maximum likelihood variance estimator (if
noise is gaussian). For the implementation, see
``asv.step_detect.solve_potts_autogamma``.

What follows is a handwaving plausibility argument why such an
objective function makes sense, and how to end up with :math:`l_1`
rather than gaussians. Better approaches are probably to be found in
step detection literature.  If you have a better formulation,
contributions/corrections are welcome!

We assume a Bayesian model:

.. math::
   :label: prob-model

   P(\{y_i\}_{i=1}^m|\sigma,k,\{\mu_i\}_{i=1}^k,\{j_i\}_{i=1}^{k-1})
   =
   N
   \sigma^{-m}
   \exp(
   -\sigma^{-1}\sum_{r=1}^k\sum_{i=j_{r-1}+1}^{j_r} w_i |y_i - \mu_r|
   )

Here, :math:`y_i` are the :math:`m` data points at hand, :math:`k` is
the number of intervals, :math:`\mu_i` are the values of the function
at the intervals, :math:`j_i` are the interval breakpoints;
:math:`j_0=0`, :math:`j_k=m`, :math:`j_{r-1}<j_r`. The noise is
assumed Laplace rather than gaussian, which results to the more
robust :math:`l_1` norm fitting rather than :math:`l_2`.  The noise
amplitude :math:`\sigma` is not known.
:math:`N` is a normalization constant that depends on :math:`m` but
not on the other parameters.

The optimal :math:`k` comes from Bayesian reasoning:
:math:`\hat{k} = \mathop{\mathrm{argmax}}_k P(k|\{y\})`, where

.. math::
   :label: p-k-prob

   P(k|\{y\}) = \frac{\pi(k)}{\pi(\{y\})}\int d\sigma (d\mu)^k \sum_{\{j\}}
   P(\{y\}|\sigma,k,\{\mu\},\{j\}) \pi(\sigma, \{\mu\},\{j\}|k)

The prior :math:`\pi(\{y\})` does not matter for :math:`\hat{k}`; the
other priors are assumed flat. We would need to estimate the behavior
of the integral in the limit :math:`m\to\infty`.  We do not succeed in
doing this rigorously here, although it might be done in the literature.

Consider first saddle-point integration over :math:`\{\mu\}`,
expanding around the max-likelihood values :math:`\mu_r^*`.
The max-likelihood estimates are the weighted medians of the data points in each interval.
Change in the exponent when :math:`\mu` is perturbed is

.. math::
   :label:

   \Delta = -\sigma^{-1}\sum_{r=1}^k \sum_{i=j_{r-1}+1}^{j_r}w_i[|y_i-\mu^*_r - \delta\mu_r| - |y_i-\mu^*_r|]

Note that :math:`\sum_{i=j_{r-1}+1}^{j_r}
w_i\mathrm{sgn}(y_i-\mu^*_r)=0`, so that response to small variations
:math:`\delta\mu_r` is :math:`m`-independent. For larger variations, we have

.. math::
   :label:

   \Delta = -\sigma^{-1}\sum_{r=1}^k N_r(\delta \mu_r) |\delta \mu_r|

where :math:`N_r(\delta\mu)=\sum_{i} w_i s_i` where :math:`s_i = \pm1` depending on whether
:math:`y_i` is above or below the perturbed median. Let us assume that in a typical case,
:math:`N_r(\delta\mu)\sim{}m_r\bar{W}_r^2\delta\mu/\sigma` where :math:`\bar{W}_r = \frac{1}{m_r}\sum_i w_i` is
the average weight of the interval and :math:`m_r` the number of points in the interval.
This recovers a result we would
have obtained in the gaussian noise case

.. math::
   :label:

   \Delta \sim -\sigma^{-2} \sum_r W_r^2 m_r |\delta \mu_r|^2

For the gaussian case, this would not have required any questionable assumptions.
After integration over :math:`\{\delta\mu\}` we are left with

.. math::
   :label:

   \int(\ldots)
   \propto
   \int d\sigma \sum_{\{j\}}
   (2\pi)^{k/2}\sigma^k [\bar{W}_1\cdots \bar{W}_k]^{-1} [m_1\cdots m_k]^{-1/2}
   P(\{y\}|\sigma,k,\{\mu_*\},\{j\})
   \pi(\sigma, \{j\}|k)

We now approximate the rest of the integrals/sums with only the
max-likelihood terms, and assume :math:`m_j^*\sim{}m/k`. Then,

.. math::
   :label: p-k-estimate

   \ln P(k|\{y\})
   &\simeq
   C_1(m) + C_2(k)
   +
   \frac{k}{2}\ln(2\pi) - \frac{k}{2} \ln(m/k) - k \ln \bar{W}
   +
   \ln P(\{y\}|\sigma_*,k,\{\mu_*\},\{j_*\})
   \\
   &\approx
   \tilde{C}_1(m) + \tilde{C}_2(k)
   -
   \frac{k}{2} \ln m
   +
   \ln P(\{y\}|\sigma_*,k,\{\mu_*\},\{j_*\})

where we neglect terms that don't affect asymptotics for
:math:`m\to\infty`, and :math:`C` are some constants not depending on
both :math:`m, k`. The result is of course the Schwarz criterion for
:math:`k` free model parameters. We can suspect that the factor
:math:`k/2` should be replaced by a different number, since we have
:math:`2k` parameters. If also the other integrals/sums can be 
approximated in the same way as the :math:`\{\mu\}` ones, we should
obtain the missing part.

Substituting in the max-likelihood value

.. math::
   :label:

    \sigma_* = \frac{1}{m} \sum_{r=1}^k\sum_{i=j_{r-1}^*+1}^{j_r^*} w_i|y_i - \mu_r^*|

we get

.. math::
   :label:

   \ln P(k|\{y\})
   \sim
   C
   -
   \frac{k}{2} \ln m
   -
   m \ln\sum_{r=1}^k\sum_{i=j_{r-1}^*}^{j_r^*} w_i |y_i - \mu_r^*|

This is now similar to :eq:`ic-form`, apart from numerical prefactors.
The final fitting problem then becomes

.. math::
   :label: bic-form

   \mathop{\mathrm{argmin}}_{k,\{j\},\{\mu\}} r(m) k + \ln\sum_{r=1}^k\sum_{i=j_{r-1}}^{j_r} w_i |y_i - \mu_r|

with :math:`r(m) = \frac{\ln m}{2m}`.
Note that it is invariant vs. rescaling of weights :math:`w_i\mapsto{}\alpha{}w_i`,
i.e., the invariance of the original problem is retained.
As we know this function
:math:`r(m)` is not necessarily completely correct, and it seems doing
the calculation rigorously requires more effort than can be justified
by the requirements of the application, we now take a pragmatic view and
fudge the function to :math:`r(m) = \beta \frac{\ln m}{m}` with
:math:`\beta` chosen so that things appear to work in practice
for the problem at hand.

According to [Friedrich2008]_, problem :eq:`bic-form` can be solved in
:math:`{\cal O}(n^3)` time.  This is too slow, however. We can however
approach this on the basis of the easier problem :eq:`gamma-opt`.  It
produces a family of solutions :math:`[k^*(\gamma), \{\mu^*(\gamma)\},
\{j^*(\gamma)\}]`.  We now evaluate :eq:`bic-form` restricted to the
curve parameterized by :math:`\gamma`.  In particular,
:math:`[\{\mu^*(\gamma)\}, \{j^*(\gamma)\}]` solves :eq:`bic-form`
under the constraint :math:`k=k^*(\gamma)`. If :math:`k^*(\gamma)`
obtains all values in the set :math:`\{1,\ldots,m\}` when
:math:`\gamma` is varied, the original problem is solved
completely. This probably is not a far-fetched assumption; in practice
it appears such Bayesian information criterion provides a reasonable
way for selecting a suitable :math:`\gamma`.


Overfitting
-----------

It's possible to fit any data perfectly by choosing size-1 intervals,
one per each data point.  For such a fit, the logarithm :eq:`bic-form`
gives :math:`-\infty` which then always minimizes SC.  This artifact
of the model above needs special handling.

Indeed, for :math:`\sigma\to0`, :eq:`prob-model` reduces to

.. math::
   :label: prob-model-2

   P(\{y_i\}_{i=1}^m|\sigma,k,\{\mu_i\}_{i=1}^k,\{j_i\}_{i=1}^{k-1})
   =
   \prod_{r=1}^k \prod_{i=j_{r-1} + 1}^{j_r} \delta(y_i - \mu_r)

which in :eq:`p-k-prob` gives a contribution (assuming no repeated y-values)

.. math::
   :label: p-k-prob-2

   P(k|\{y\}) = \frac{\pi(n)}{\pi(\{y\})}\delta_{n,k}\int d\sigma
   \pi(\sigma, \{y\},\{i\}|n) f(\sigma)
   +
   \ldots

with :math:`f(\sigma)\to1` for :math:`\sigma\to0`.  A similar situation
occurs also in other cases where perfect fitting occurs (repeated
y-values).  With the flat, scale-free prior
:math:`\pi(\ldots)\propto1/\sigma` used above, the result is
undefined.

A simple fix is to give up complete scale free-ness of the results,
i.e., fixing a minimal noise level
:math:`\pi(\sigma,\{\mu\},\{j\}|k)\propto\theta(\sigma-\sigma_0)/\sigma` with some
:math:`\sigma_0(\{\mu\},\{j\},k)>0`. The effect in the
:math:`\sigma` integral is cutting off the log-divergence, so that
with sufficient accuracy we can in :eq:`bic-form` replace

.. math::
   :label: bic-form-2

   \ln \sigma \mapsto \ln(\sigma_0 + \sigma)

Here, we fix a measurement accuracy floor with the following guess:
``sigma_0 = 0.1 * w0 * min(abs(diff(mu)))`` and ``sigma_0 = 0.001 * w0
* abs(mu)`` when there is only a single interval. Here, ``w0`` is the
median weight.

Autocorrelated noise
--------------------

Practical experience shows that the noise in the benchmark results can be
correlated. Often benchmarks are run for multiple commits at once, for
example the new commits at a given time, and the benchmark machine
does something else between the runs. Alternatively, the background
load from other processes on the machine varies with time.

To give a basic model for the noise correlations, we include
AR(1) Laplace noise in :eq:`prob-model`,

.. math::
   :label: autocorr-model

   P(\{y_i\}_{i=1}^m|\sigma,\rho,k,\{\mu_i\}_{i=1}^k,\{j_i\}_{i=1}^{k-1})
   =
   N
   \sigma^{-m}
   \exp(-\sigma^{-1}\sum_{r=1}^k\sum_{i=j_{r-1}+1}^{j_r} |\epsilon_{i,r} - \rho \epsilon_{i-1,r}|)

where :math:`\epsilon_{i,r}=y_i-\mu_{r}` with
:math:`\epsilon_{j_{r-1},r}=y_{j_{r-1}}-\mu_{r-1}` and
:math:`\epsilon_{j_0,1}=0` are the deviations from the stepwise
model. The correlation measure :math:`\rho` is unknown, but assumed to
be constant in :math:`(-1,1)`.

Since the parameter :math:`\rho` is global, it does not change the parameter
counting part of the Schwarz criterion.  The maximum likelihood term however
does depend on :math:`\rho`, so that the problem becomes:

.. math::
   :label: bic-form-autocorr

   \mathop{\mathrm{argmin}}_{k,\rho,\{j\},\{\mu\}} r(m) k + \ln\sum_{r=1}^k\sum_{i=j_{r-1}}^{j_r} |\epsilon_{i,r} - \rho\epsilon_{i-1,r}|

To save computation time, we do not solve this optimization problem
exactly. Instead, we again minimize along the :math:`\mu_r^*(\gamma)`,
:math:`j_r^*(\gamma)` curve provided by the solution to
:eq:`gamma-opt`, and use :eq:`bic-form-autocorr` only in selecting the
optimal value of the :math:`\gamma` parameter.

The minimization vs. :math:`\rho` can be done numerically for given
:math:`\mu_r^*(\gamma)`, :math:`j_r^*(\gamma)`. This minimization step
is computationally cheap compared to the piecewise fit, so including
it will not significantly change the runtime of the total algorithm.

Postprocessing
--------------

For the purposes of regression detection, we do not report all steps
the above approach provides. For details, see
``asv.step_detect.detect_regressions``.

Making use of measured variance
-------------------------------

``asv`` measures also variance in the timings.  This information is
currently used to provide relative data weighting (see above).

"""

from __future__ import absolute_import, division, unicode_literals, print_function

import math
import collections
import heapq
import six

try:
    from . import _rangemedian
except ImportError:
    _rangemedian = None


#
# Detecting regressions
#

def detect_steps(y, w=None):
    """
    Detect steps in a (noisy) signal.

    Parameters
    ----------
    y : list of float, none or nan
        Single benchmark result series, with possible missing data
    w : list of float, none or nan
        Data point relative weights. Missing weights are set equal
        to the median weight.

    Returns
    -------
    steps : list of (left_pos, right_pos, value, min_value, err_est)
        List containing a decomposition of the input data to a piecewise
        constant function. Each element contains the left (inclusive) and
        right (exclusive) bounds of a segment, the average value on 
        the segment, the minimum value in the segment, and the l1 error 
        estimate, <|Y - avg|>. Missing data points are not necessarily
        contained in any segment; right_pos-1 is the last non-missing data
        point.

    """

    index_map = {}
    y_filtered = []
    for j, x in enumerate(y):
        if x is None or x != x:
            # None or NaN: missing data
            continue
        if w is not None and w[j] is not None and (w[j] <= 0):
            # non-positive weight: consider as missing data
            continue
        index_map[len(y_filtered)] = j
        y_filtered.append(x)

    # Weights
    if w is None:
        w_filtered = [1]*len(y_filtered)
    else:
        # Fill-in and normalize weights
        w_valid = [ww for ww in w if ww is not None and ww == ww]
        if w_valid:
            w_median = median(w_valid)
            if w_median == 0:
                w_median = 1.0
        else:
            w_median = 1.0

        w_filtered = [1.0]*len(y_filtered)
        for j in range(len(w_filtered)):
            jj = index_map[j]
            if w[jj] is not None and w[jj] == w[jj]:
                w_filtered[j] = w[jj] / w_median

    # Find piecewise segments
    right, values, dists, gamma = solve_potts_autogamma(y_filtered, w=w_filtered)

    # Extract the steps, mapping indices back etc.
    steps = []
    l = 0
    for r, v, d in zip(right, values, dists):
        steps.append((index_map[l], index_map[r-1] + 1,
                          v,
                          min(y_filtered[l:r]),
                          abs(d/(r - l))))
        l = r
    return steps


def detect_regressions(steps, threshold=0, min_size=2):
    """Detect regressions in a (noisy) signal.

    A regression means an upward step in the signal.  The value
    'before' a regression is the value immediately preceding the
    upward step.  The value 'after' a regression is the minimum of
    values after the upward step.

    Parameters
    ----------
    steps : list of (left, right, value, min, error)
        List of steps computed by detect_steps, or equivalent
    threshold : float
        Relative threshold for reporting regressions. Filter out jumps
        whose relative size is smaller than threshold, if they are not
        necessary to explain the difference between the best and the latest
        values.
    min_size : int
        Minimum number of commits in a regression to consider it.

    Returns
    -------
    latest_value
        Latest value
    best_value
        Best value
    regression_pos : list of (before, after, value_before, best_value_after)
        List of positions between which the value increased. The first item
        corresponds to the last position at which the best value was obtained.
        The last item indicates the best value found after the regression
        (which is not always the value immediately following the regression).

    """
    if not steps:
        # No data: no regressions
        return None, None, None

    regression_pos = []

    last_v = steps[-1][2]
    best_v = last_v
    thresholded_best_v = last_v
    thresholded_best_err = steps[-1][4]
    prev_l = None
    short_prev = None

    # Find upward steps that resulted to worsened value afterward
    for l, r, cur_v, cur_min, cur_err in reversed(steps):
        threshold_step = max(cur_err, thresholded_best_err, threshold * cur_v)

        if thresholded_best_v > cur_v + threshold_step:
            if r - l < min_size:
                # Accept short intervals conditionally
                short_prev = (thresholded_best_v, thresholded_best_err)

            regression_pos.append((r - 1, prev_l, cur_v, best_v))

            thresholded_best_v = cur_v
            thresholded_best_err = cur_err
        elif short_prev is not None:
            # Ignore the previous short interval, if the level
            # is now back to where it was
            if short_prev[0] <= cur_v + threshold_step:
                regression_pos.pop()
                thresholded_best_v, thresholded_best_err = short_prev
            short_prev = None

        prev_l = l

        if cur_v < best_v:
            best_v = cur_v

    regression_pos.reverse()

    # Return results
    if regression_pos:
        return (last_v, best_v, regression_pos)
    else:
        return (None, None, None)


#
# Fitting piecewise constant functions to noisy data
#

def solve_potts(y, w, gamma, min_size=1, max_size=None,
                min_pos=None, max_pos=None, mu_dist=None):
    """Fit penalized stepwise constant function (Potts model) to data.

    Given a time series y = {y_1, ..., y_n}, fit series x = {x_1, ..., x_n}
    by minimizing the cost functional::

        F[x] = gamma * J(x) + sum(|y - x|**p)

    where J(x) is the number of jumps (x_{j+1} != x_j) in x.

    The algorithm used is described in Ref. [1]_, it uses dynamic
    programming to find an exact solution to the problem (within the
    constraints specified).

    Computation work is ~ O(n**2 log n).

    Parameters
    ----------
    y : list of floats
        Input data series
    gamma : float
        Penalty parameter.
    min_size : int, optional
        Minimum interval size to consider
    max_size : int, optional
        Maximum interval size to consider
    mu_dist : *Dist, optional
        Precomputed interval means/medians and cost function values
    min_pos : int, optional
        Start point (inclusive) for the interval grid
    max_pos : int, optional
        End point (exclusive) for the interval grid

    Returns
    -------
    right : list
        List of (exclusive) right bounds of the intervals
    values : list
        List of values of the intervals
    dist : list
        List of ``sum(|y - x|**p)`` for each interval.
    mu_dist : *Dist
        Precomputed interval means/medians and cost function values

    References
    ----------
    [1] F. Friedrich et al., "Complexity Penalized M-Estimation: Fast Computation",
        Journal of Computational and Graphical Statistics 17.1, 201-224 (2008).

    """

    inf = float('inf')

    if len(y) == 0:
        return [], [], []

    if min_pos is None:
        min_pos = 0

    if len(y) != len(w):
        raise ValueError("y and w must have same size")

    if max_pos is None:
        max_pos = len(y)

    if mu_dist is None:
        mu_dist = get_mu_dist(y, w)

    if max_size is None:
        max_size = len(y)

    mu, dist = mu_dist.mu, mu_dist.dist

    if min_size >= max_pos - min_pos:
        return [len(y)], [mu(0,len(y)-1)], [dist(0,len(y)-1)]

    # Perform the Bellman recursion for the optimal partition.
    # Routine "Find best partition" in [1]
    #
    # Computes:
    #
    # p : list, length n
    #     Set of intervals, represented as follows:
    #     For interval (inclusive) right edge r in {0, ..., n-1},
    #     the best (exclusive) left edge is at l=p[r].
    #     Where intervals overlap, the rightmost one has priority.

    if hasattr(mu_dist, 'find_best_partition'):
        p = mu_dist.find_best_partition(gamma, min_size, max_size, min_pos, max_pos)
    else:
        i0 = min_pos
        i1 = max_pos

        B = [-gamma]*(i1 - i0 + 1)
        p = [0]*(i1 - i0)
        for r in range(i0, i1):
            B[r+1-i0] = inf
            a = max(r + 1 - max_size, i0)
            b = max(r + 1 - min_size + 1, i0)
            for l in range(a, b):
                b = B[l-i0] + gamma + dist(l, r)
                if b <= B[r+1-i0]:
                    B[r+1-i0] = b
                    p[r-i0] = l - 1

            mu_dist.cleanup_cache()

    # Routine "Segmentation from partition" in [1]
    # Convert interval representation computed above
    # to a list of intervals and values.
    r = len(p) - 1 + min_pos
    l = p[r - min_pos]
    right = []
    values = []
    dists = []
    while r >= min_pos:
        right.append(r + 1)
        values.append(mu((l + 1), r))
        dists.append(dist((l + 1), r))
        r = l
        l = p[r - min_pos]
    right.reverse()
    values.reverse()
    dists.reverse()

    return right, values, dists


def solve_potts_autogamma(y, w, beta=None, **kw):
    """Solve Potts problem with automatically determined gamma.

    The optimal value is determined by minimizing the information measure::

        f(gamma) = beta J(x(gamma)) + log sum(abs(x(gamma) - y)**p)

    where x(gamma) is the solution to the Potts problem for a fixed
    gamma. The minimization is only performed rather roughly.

    Parameters
    ----------
    beta : float or 'bic'
         Penalty parameter. Default is 4*ln(n)/n, similar to Bayesian
         information criterion for gaussian model with unknown variance
         assuming 4 DOF per breakpoint.

    """
    n = len(y)

    if n == 0:
        return [], [], [], None

    mu_dist = get_mu_dist(y, w)
    mu, dist = mu_dist.mu, mu_dist.dist

    if beta is None:
        beta = 4 * math.log(n) / n

    gamma_0 = dist(0, n-1)

    if gamma_0 == 0:
        # Zero variance
        gamma_0 = 1.0

    best_r = [None]
    best_v = [None]
    best_d = [None]
    best_obj = [float('inf')]
    best_gamma = [None]

    def f(x):
        gamma = gamma_0 * math.exp(x)
        r, v, d = solve_potts_approx(y, w, gamma=gamma, mu_dist=mu_dist, **kw)

        # MLE fit noise correlation
        def sigma_star(rights, values, rho):
            """
            |E_0| + sum_{j>0} |E_j - rho E_{j-1}|
            """
            l = 1
            E_prev = y[0] - values[0]
            s = abs(E_prev) * w[0]
            for r, v in zip(rights, values):
                for yv, wv in zip(y[l:r], w[l:r]):
                    E = yv - v
                    s += abs(E - rho*E_prev) * wv
                    E_prev = E
                l = r
            return s

        rho_best = golden_search(lambda rho: sigma_star(r, v, rho), -1, 1,
                                 xatol=0.05, expand_bounds=True)

        # Measurement noise floor
        if len(v) > 2:
            absdiff = [abs(v[j+1] - v[j]) for j in range(len(v) - 1)]
            sigma_0 = 0.1 * min(absdiff)
        else:
            absv = [abs(z) for z in v]
            sigma_0 = 0.001 * min(absv)
        sigma_0 = max(1e-300, sigma_0)

        # Objective function
        s = sigma_star(r, v, rho_best)
        obj = beta*len(r) + math.log(sigma_0 + s)

        # Done
        if obj < best_obj[0]:
            best_r[0] = r
            best_v[0] = v
            best_d[0] = d
            best_gamma[0] = gamma
            best_obj[0] = obj
        return obj

    # Try to find best gamma (golden section search on log-scale); we
    # don't need an accurate value for it however
    a = math.log(0.1/n)
    b = 0.0
    golden_search(f, a, b, xatol=abs(a)*0.1, ftol=0, expand_bounds=True)
    return best_r[0], best_v[0], best_d[0], best_gamma[0]


def solve_potts_approx(y, w, gamma=None, min_size=1, **kw):
    """
    Fit penalized stepwise constant function (Potts model) to data
    approximatively, in linear time.

    Do this by running the exact solver using a small maximum interval
    size, and then combining consecutive intervals together if it
    decreases the cost function.
    """
    n = len(y)

    if n == 0:
        return [], [], []

    mu_dist = kw.get('mu_dist')
    if mu_dist is None:
        mu_dist = get_mu_dist(y, w)
        kw['mu_dist'] = mu_dist

    if gamma is None:
        mu, dist = mu_dist.mu, mu_dist.dist
        gamma = 3 * dist(0,n-1) * math.log(n) / n

    if min_size < 10:
        max_size = 20
    else:
        max_size = min_size + 50

    right, values, dists = solve_potts(y, w, gamma, min_size=min_size, max_size=max_size, **kw)
    return merge_pieces(gamma, right, values, dists, mu_dist, max_size=max_size)


def merge_pieces(gamma, right, values, dists, mu_dist, max_size):
    """
    Combine consecutive intervals in Potts model solution, if doing
    that reduces the cost function.
    """
    mu, dist = mu_dist.mu, mu_dist.dist

    right = list(right)

    # Combine consecutive intervals, if it results to decrease of cost
    # function
    while True:
        min_change = 0
        min_change_j = len(right)

        l = 0
        for j in range(1, len(right)):
            if min_change_j < j - 2:
                break

            # Check whether merging consecutive intervals results to
            # decrease in the cost function
            change = dist(l, right[j]-1) - (dist(l, right[j-1]-1) + dist(right[j-1], right[j]-1) + gamma)
            if change <= min_change:
                min_change = change
                min_change_j = j-1
            l = right[j-1]

        if min_change_j < len(right):
            del right[min_change_j]
        else:
            break

    # Check whether perturbing boundary positions leads to improvement
    # in the cost function. The restricted Potts minimization can
    # return sub-optimal boundaries due to the interval maximum size
    # restriction.
    l = 0
    for j in range(1, len(right)):
        prev_score = dist(l, right[j-1]-1) + dist(right[j-1], right[j]-1)
        new_off = 0
        for off in range(-max_size, max_size+1):
            if right[j-1] + off - 1 < l or right[j-1] + off > right[j] - 1 or off == 0:
                continue
            new_score = dist(l, right[j-1]+off-1) + dist(right[j-1]+off, right[j]-1)
            if new_score < prev_score:
                new_off = off
                prev_score = new_score

        if new_off != 0:
            right[j-1] += new_off

        l = right[j-1]

    # Rebuild values and dists lists
    l = 0
    values = []
    dists = []
    for j in range(len(right)):
        dists.append(dist(l, right[j]-1))
        values.append(mu(l, right[j]-1))
        l = right[j]

    return right, values, dists


class L1Dist(object):
    """
    Fast computations for::

        mu(l, r) = median(y[l:r+1], weights=w[l:r+1])
        dist(l, r) = sum(w*abs(x - mu(l, r)) for x, w in zip(y[l:r+1], weights[l:r+1]))

    We do not use here an approach that has asymptotically optimal
    performance; at least O(n**2 * log(n)) would be achievable, whereas
    we have here O(n**3).  The asymptotic performance does not matter
    for solve_potts_approx, which only looks at small windows of the
    data. It is more important to try to optimize the constant
    prefactors, which for Python means minimal code.

    """
    def __init__(self, y, w):
        self.y = y
        self.w = w

        class mu_dict(collections.defaultdict):
            def __missing__(self, a):
                l, r = a
                v = weighted_median(y[l:r+1], w[l:r+1])
                self[a] = v
                return v

        mu = mu_dict()

        class dist_dict(collections.defaultdict):
            def __missing__(self, a):
                l, r = a
                m = mu[l, r]
                v = sum(wx*abs(x - m) for x, wx in zip(y[l:r+1], w[l:r+1]))
                self[a] = v
                return v

        self.mu_memo = mu
        self.dist_memo = dist_dict()

    def mu(self, *a):
        return self.mu_memo[a]

    def dist(self, *a):
        return self.dist_memo[a]

    def cleanup_cache(self):
        # Reset cache if it is too big
        if len(self.mu_memo) < 500000:
            return

        self.mu_memo.clear()
        self.dist_memo.clear()


def get_mu_dist(y, w):
    if _rangemedian is not None:
        return _rangemedian.RangeMedian(y, w)
    else:
        return L1Dist(y, w)


def median(items):
    """Note: modifies the input list!"""
    items.sort()
    k = len(items)//2
    if len(items) % 2 == 0:
        return (items[k] + items[k - 1]) / 2
    else:
        return items[k]


def rolling_median_dev(items):
    """
    Compute median(items[:j]), deviation[j]) for j in range(1, len(items))
    in O(n log n) time.

    deviation[j] == sum(abs(x - median(items[:j])) for x in items[:j])
    """
    min_heap = []
    max_heap = []
    min_heap_sum = 0   # equal to -sum(min_heap)
    max_heap_sum = 0   # equal to sum(max_heap)
    s = iter(items)
    try:
        while True:
            # Odd
            v = six.next(s)
            min_heap_sum += v
            v = -heapq.heappushpop(min_heap, -v)
            min_heap_sum -= v
            heapq.heappush(max_heap, v)
            max_heap_sum += v
            # Ensure d >= 0 despite rounding error
            d = max(0, max_heap_sum - min_heap_sum - max_heap[0])
            yield (max_heap[0], d)

            # Even
            v = six.next(s)
            max_heap_sum += v
            v = heapq.heappushpop(max_heap, v)
            max_heap_sum -= v
            heapq.heappush(min_heap, -v)
            min_heap_sum += v
            d = max(0, max_heap_sum - min_heap_sum)
            yield ((max_heap[0] - min_heap[0])/2, d)
    except StopIteration:
        return


def weighted_median(y, w):
    """
    Compute weighted median of `y` with weights `w`.
    """
    items = sorted(zip(y, w))
    midpoint = sum(w) / 2

    yvals = []
    wsum = 0

    for yy, ww in items:
        wsum += ww
        if wsum > midpoint:
            yvals.append(yy)
            break
        elif wsum == midpoint:
            yvals.append(yy)
    else:
        yvals = y

    return sum(yvals) / len(yvals)


def golden_search(f, a, b, xatol=1e-6, ftol=1e-8, expand_bounds=False):
    """
    Find minimum of a function on interval [a, b]
    using golden section search.

    If expand_bounds=True, expand the interval so that the function is
    first evaluated at x=a and x=b.
    """

    ratio = 2 / (1 + math.sqrt(5))

    if not expand_bounds:
        x0 = a
        x3 = b
    else:
        x0 = (ratio * a - (1 - ratio) * b) / (2*ratio - 1)
        x3 = (ratio * b - (1 - ratio) * a) / (2*ratio - 1)

    x1 = ratio * x0 + (1 - ratio) * x3
    x2 = (1 - ratio) * x0 + ratio * x3

    f1 = f(x1)
    f2 = f(x2)

    f0 = max(abs(f1), abs(f2))

    while True:
        if abs(x0 - x3) < xatol or abs(f1 - f2) < ftol*f0:
            break

        if f2 < f1:
            x0 = x1
            x1 = x2
            x2 = ratio * x1 + (1 - ratio) * x3
            f1 = f2
            f2 = f(x2)
        else:
            x3 = x2
            x2 = x1
            x1 = ratio * x2 + (1 - ratio) * x0
            f2 = f1
            f1 = f(x1)

    if f2 < f1:
        return x2
    else:
        return x1


def _plot_potts(x, sol):
    import numpy as np
    import matplotlib.pyplot as plt

    t = np.arange(len(x))

    plt.clf()
    plt.plot(t, x, 'k.')

    l = 0
    for r, v in zip(sol[0], sol[1]):
        plt.plot([l, r-1], [v, v], 'b-o', hold=1)
        l = r



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Author: Pauli Virtanen, 2016

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import math
import operator

from .util import inf, nan, is_na


def compute_stats(samples, number):
    """
    Statistical analysis of measured samples.

    Parameters
    ----------
    samples : list of float
        List of total times (y) of benchmarks.
    number : int
        Repeat number for each sample.

    Returns
    -------
    beta_hat : float
        Estimated time per iteration
    stats : dict
        Information on statistics of the estimator.

    """

    if len(samples) < 1:
        return None, None

    Y = list(samples)

    # Median and quantiles
    y_50, ci_50 = quantile_ci(Y, 0.5, alpha_min=0.99)
    y_25 = quantile(Y, 0.25)
    y_75 = quantile(Y, 0.75)

    # If nonparametric CI estimation didn't give an estimate,
    # use the credible interval of a bayesian posterior distribution.
    a, b = ci_50
    if (math.isinf(a) or math.isinf(b)) and len(Y) > 1:
        # Compute posterior distribution for location, assuming
        # exponential noise. The MLE is equal to the median.
        c = LaplacePosterior(Y)

        # Use the CI from that distribution to extend beyond sample
        # bounds
        if math.isinf(a):
            a = min(c.ppf(0.01/2), min(Y))
        if math.isinf(b):
            b = max(c.ppf(1 - 0.01/2), max(Y))

        ci_50 = (a, b)

    # Produce results
    result = y_50

    stats = {'ci_99_a': ci_50[0],
             'ci_99_b': ci_50[1],
             'q_25': y_25,
             'q_75': y_75,
             'repeat': len(Y),
             'number': number}

    return result, stats


def get_err(result, stats):
    """
    Return an 'error measure' suitable for informing the user
    about the spread of the measurement results.
    """
    a, b = stats['q_25'], stats['q_75']
    return (b - a)/2


def get_weight(stats):
    """
    Return a data point weight for the result.
    """
    if stats is None or 'ci_99_a' not in stats or 'ci_99_b' not in stats:
        return None

    try:
        a = stats['ci_99_a']
        b = stats['ci_99_b']

        if math.isinf(a) or math.isinf(b):
            # Infinite interval is due to too few samples --- consider
            # weight as missing
            return None

        return 2 / abs(b - a)
    except ZeroDivisionError:
        return None


def is_different(samples_a, samples_b, stats_a, stats_b, p_threshold=0.002):
    """Check whether the samples are statistically different.

    If sample data is not provided, or the sample is too small, falls
    back to a pessimistic CI-based check. If it returns True, then the
    difference is statistically significant. If it returns False, it
    might or might not be statistically significant.

    Parameters
    ----------
    samples_a, samples_b
        Input samples
    stats_a, stats_b
        Input stats data

    """

    if samples_a is not None and samples_b is not None:
        # Raw data present: Mann-Whitney U test, but only if there's
        # enough data so that the test can return True
        a = [x for x in samples_a if not is_na(x)]
        b = [x for x in samples_b if not is_na(x)]

        p_min = 1 / binom(len(a) + len(b), min(len(a), len(b)))
        if p_min < p_threshold:
            _, p = mann_whitney_u(a, b)
            return p < p_threshold

    # If confidence intervals overlap, reject.
    # Corresponds to a test with ill-specified threshold p-value,
    # which generally can be significantly smaller than p <= 0.01
    # depending on the actual data. For normal test (known variance),
    # 0.00027 <= p <= 0.01.
    ci_a = (stats_a['ci_99_a'], stats_a['ci_99_b'])
    ci_b = (stats_b['ci_99_a'], stats_b['ci_99_b'])

    if ci_a[1] >= ci_b[0] and ci_a[0] <= ci_b[1]:
        return False

    return True


def quantile_ci(x, q, alpha_min=0.01):
    """
    Compute a quantile and a confidence interval.

    Assumes independence, but otherwise nonparametric.

    Parameters
    ----------
    x : list of float
        Samples
    q : float
        Quantile to compute, in [0,1].
    alpha_min : float, optional
        Limit for coverage.
        The result has coverage >= 1 - alpha_min.

    Returns
    -------
    m : float
        Quantile of x
    ci : tuple of floats
        Confidence interval (a, b), of coverage >= alpha_min.

    """

    y = sorted(x)
    n = len(y)

    alpha_min = min(alpha_min, 1 - alpha_min)
    pa = alpha_min / 2
    pb = 1 - pa

    a = -inf
    b = inf

    # It's known that
    #
    # Pr[X_{(r)} < m < X_{(s)}] = Pr[r <= K <= s-1], K ~ Bin(n,p)
    #
    # where cdf(m) = p defines the quantile.
    #
    # Simplest median CI follows by picking r,s such that
    #
    #   F(r;n,q) <= alpha/2
    #   F(s;n,q) >= 1 - alpha/2
    #
    #   F(k;n,q) = sum(binom_pmf(n, j, q) for j in range(k))
    #
    # Then (y[r-1], y[s-1]) is a CI.
    # If no such r or s exists, replace by +-inf.

    F = 0
    for k, yp in enumerate(y):
        F += binom_pmf(n, k, q)
        # F = F(k+1;n,q)

        if F <= pa:
            a = yp

        if F >= pb:
            b = yp
            break

    m = quantile(y, q)
    return m, (a, b)


def quantile(x, q):
    """
    Compute quantile/percentile of the data

    Parameters
    ----------
    x : list of float
        Data set
    q : float
        Quantile to compute, 0 <= q <= 1

    """
    if not 0 <= q <= 1:
        raise ValueError("Invalid quantile")

    y = sorted(x)
    n = len(y)

    z = (n - 1) * q
    j = int(math.floor(z))
    z -= j

    if j == n - 1:
        m = y[-1]
    else:
        m = (1 - z)*y[j] + z*y[j+1]

    return m


_mann_whitney_u_memo = {}

def mann_whitney_u(x, y, method='auto'):
    """
    Mann-Whitney U test

    Ties are handled conservatively, returning the least significant
    tie breaking.

    Parameters
    ----------
    x, y : list of float
        Samples to test
    method : {'auto', 'exact', 'normal'}
        Whether to compute p-value exactly of via normal approximation.
        The option 'auto' switches to approximation for sample size > 20.

    Returns
    -------
    u : int
        U-statistic
    p : float
        p-value for two-sided alternative

    References
    ----------
    .. [1] Mann & Whitney, Ann. Math. Statist. 18, 50 (1947).
    .. [2] Gibbons & Chakraborti, "Nonparametric statistical inference". (2003)

    """
    memo = _mann_whitney_u_memo
    if len(memo) > 100000:
        memo.clear()

    m = len(x)
    n = len(y)

    if method == 'auto':
        if max(m, n) > 20:
            method = 'normal'
        else:
            method = 'exact'

    u, ties = mann_whitney_u_u(x, y)

    # Conservative tie breaking
    if u <= m*n//2 and u + ties >= m*n//2:
        ties = m*n//2 - u

    ux1 = min(u, m*n - u)
    ux2 = min(u + ties, m*n - (u + ties))

    if ux1 >= ux2:
        ux = ux1
    else:
        u = u + ties
        ux = ux2

    # Get p-value
    if method == 'exact':
        p1 = mann_whitney_u_cdf(m, n, ux, memo)
        p2 = 1.0 - mann_whitney_u_cdf(m, n, max(m*n//2, m*n - ux - 1), memo)
        p = p1 + p2
    elif method == 'normal':
        N = m + n
        var = m*n*(N + 1) / 12
        z = (ux - m*n/2) / math.sqrt(var)
        cdf = 0.5 * math.erfc(-z / math.sqrt(2))
        p = 2 * cdf
    else:
        raise ValueError("Unknown method {!r}".format(method))

    return u, p


def mann_whitney_u_u(x, y):
    u = 0
    ties = 0
    for xx in x:
        for yy in y:
            if xx > yy:
                u += 1
            elif xx == yy:
                ties += 1
    return u, ties


def mann_whitney_u_cdf(m, n, u, memo=None):
    if memo is None:
        memo = {}
    cdf = 0
    for uu in range(u + 1):
         cdf += mann_whitney_u_pmf(m, n, uu, memo)
    return cdf


def mann_whitney_u_pmf(m, n, u, memo=None):
    if memo is None:
        memo = {}
    return mann_whitney_u_r(m, n, u, memo) / binom(m + n, m)


def mann_whitney_u_r(m, n, u, memo=None):
    """
    Number of orderings in Mann-Whitney U test.

    The PMF of U for samples of sizes (m, n) is given by
    p(u) = r(m, n, u) / binom(m + n, m).

    References
    ----------
    .. [1] Mann & Whitney, Ann. Math. Statist. 18, 50 (1947).
    """
    if u < 0:
        value = 0
    elif m == 0 or n == 0:
        value = 1 if u == 0 else 0
    else:
        # Don't bother figuring out table construction, memoization
        # sorts it out
        if memo is None:
            memo = {}
        key = (m, n, u)
        value = memo.get(key)
        if value is not None:
            return value

        value = (mann_whitney_u_r(m, n - 1, u, memo)
                 + mann_whitney_u_r(m - 1, n, u - n, memo))

        memo[key] = value
    return value


def binom_pmf(n, k, p):
    """Binomial pmf = (n choose k) p**k (1 - p)**(n - k)"""
    if not (0 <= k <= n):
        return 0
    if p == 0:
        return 1.0 * (k == 0)
    elif p == 1.0:
        return 1.0 * (k == n)

    logp = math.log(p)
    log1mp = math.log(1 - p)
    return math.exp(lgamma(1 + n) - lgamma(1 + n - k) - lgamma(1 + k)
                        + k*logp + (n - k)*log1mp)


_BERNOULLI = [1.0, -0.5, 0.166666666667, 0.0, -0.0333333333333, 0.0, 0.0238095238095]


def lgamma(x):
    """
    Log gamma function. Only implemented at integers.
    """

    if x <= 0:
        raise ValueError("Domain error")

    if x > 100:
        # DLMF 5.11.1
        r = 0.5 * math.log(2*math.pi) + (x - 0.5) * math.log(x) - x
        for k in range(1, len(_BERNOULLI)//2 + 1):
            r += _BERNOULLI[2*k] / (2*k*(2*k - 1) * x**(2*k - 1))
        return r

    # Fall back to math.factorial
    int_x = int(x)
    err_int = abs(x - int_x)

    if err_int < 1e-12 * abs(x):
        return math.log(math.factorial(int_x - 1))

    # Would need full implementation
    return nan


def binom(n, k):
    """
    Binomial coefficient (n over k)
    """
    n = operator.index(n)
    k = operator.index(k)
    if not 0 <= k <= n:
        return 0
    m = n + 1
    num = 1
    den = 1
    for j in range(1, min(k, n - k) + 1):
        num *= m - j
        den *= j
    return num // den


class LaplacePosterior(object):
    """
    Univariate distribution::

        p(beta|y) = N [sum(|y_j - beta|)]**(-nu-1)

    where N is the normalization factor.

    Parameters
    ----------
    y : list of float
        Samples
    nu : float, optional
        Degrees of freedom. Default: len(y)-1

    Notes
    -----
    This is the posterior distribution in the Bayesian model assuming
    Laplace distributed noise::

        p(y|beta,sigma) = N exp(- sum_j (1/sigma) |y_j - beta|)

        p(sigma) ~ 1/sigma

        nu = len(y) - 1

    The MLE for beta is median(y).

    Note that the same approach applied to a Gaussian model::

        p(y|beta,sigma) = N exp(- sum_j 1/(2 sigma^2) (y_j - beta)^2)

    results to::

        p(beta|y) = N T(t, m-1);   t = (beta - mean(y)) / (sstd(y) / sqrt(m))

    where ``T(t, nu)`` is the Student t-distribution pdf, which then gives
    the standard textbook formulas for the mean.

    """

    def __init__(self, y, nu=None):
        if len(y) == 0:
            raise ValueError("empty input")

        if nu is None:
            self.nu = len(y) - 1
        else:
            self.nu = nu

        # Sort input
        y = sorted(y)

        # Get location and scale so that data is centered at MLE, and
        # the unnormalized PDF at MLE has amplitude ~ 1/nu.
        #
        # Proper scaling of inputs is important to avoid overflows
        # when computing the unnormalized CDF integrals below.
        self.mle = quantile(y, 0.5)
        self._y_scale = sum(abs(yp - self.mle) for yp in y)
        self._y_scale *= self.nu**(1/(self.nu + 1))

        # Shift and scale
        if self._y_scale != 0:
            self.y = [(yp - self.mle)/self._y_scale for yp in y]
        else:
            self.y = [0 for yp in y]

        self._cdf_norm = None
        self._cdf_memo = {}

    def _cdf_unnorm(self, beta):
        """
        Unnormalized CDF of this distribution::

            cdf_unnorm(b) = int_{-oo}^{b} 1/(sum_j |y - b'|)**(m+1) db'

        """
        if beta != beta:
            return beta

        for k, y in enumerate(self.y):
            if y > beta:
                k0 = k
                break
        else:
            k0 = len(self.y)

        cdf = 0

        nu = self.nu

        # Save some work by memoizing intermediate results
        if k0 - 1 in self._cdf_memo:
            k_start = k0
            cdf = self._cdf_memo[k0 - 1]
        else:
            k_start = 0
            cdf = 0

        # Do the integral piecewise, resolving the absolute values
        for k in range(k_start, k0 + 1):
            c = 2*k - len(self.y)
            y = sum(self.y[k:]) - sum(self.y[:k])

            if k == 0:
                a = -inf
            else:
                a = self.y[k-1]

            if k == k0:
                b = beta
            else:
                b = self.y[k]

            if c == 0:
                term = (b - a) / y**(nu+1)
            else:
                term = 1/(nu*c) * ((a*c + y)**(-nu) - (b*c + y)**(-nu))

            cdf += max(0, term)  # avoid rounding error

            if k != k0:
                self._cdf_memo[k] = cdf

        if beta == inf:
            self._cdf_memo[len(self.y)] = cdf

        return cdf

    def _ppf_unnorm(self, cdfx):
        """
        Inverse function for _cdf_unnorm
        """
        # Find interval
        for k in range(len(self.y) + 1):
            if cdfx <= self._cdf_memo[k]:
                break

        # Invert on interval
        c = 2*k - len(self.y)
        y = sum(self.y[k:]) - sum(self.y[:k])
        nu = self.nu
        if k == 0:
            term = cdfx
        else:
            a = self.y[k-1]
            term = cdfx - self._cdf_memo[k-1]

        if k == 0:
            z = -nu*c*term
            if z > 0:
                beta = (z**(-1/nu) - y) / c
            else:
                beta = -inf
        elif c == 0:
            beta = a + term * y**(nu+1)
        else:
            z = (a*c + y)**(-nu) - nu*c*term
            if z > 0:
                beta = (z**(-1/nu) - y)/c
            else:
                beta = inf

        if k < len(self.y):
            beta = min(beta, self.y[k])

        return beta

    def pdf(self, beta):
        """
        Probability distribution function
        """
        return math.exp(self.logpdf(beta))

    def logpdf(self, beta):
        """
        Logarithm of probability distribution function
        """
        if self._y_scale == 0:
            return inf if beta == self.mle else -inf

        beta = (beta - self.mle) / self._y_scale

        if self._cdf_norm is None:
            self._cdf_norm = self._cdf_unnorm(inf)

        ws = sum(abs(yp - beta) for yp in self.y)
        m = self.nu
        return -(m+1)*math.log(ws) - math.log(self._cdf_norm) - math.log(self._y_scale)

    def cdf(self, beta):
        """
        Cumulative probability distribution function
        """
        if self._y_scale == 0:
            return 1.0*(beta > self.mle)

        beta = (beta - self.mle) / self._y_scale

        if self._cdf_norm is None:
            self._cdf_norm = self._cdf_unnorm(inf)
        return self._cdf_unnorm(beta) / self._cdf_norm

    def ppf(self, cdf):
        """
        Percent point function (inverse function for cdf)
        """
        if cdf < 0 or cdf > 1.0:
            return nan

        if self._y_scale == 0:
            return self.mle

        if self._cdf_norm is None:
            self._cdf_norm = self._cdf_unnorm(inf)

        cdfx = min(cdf * self._cdf_norm, self._cdf_norm)
        beta = self._ppf_unnorm(cdfx)
        return beta * self._y_scale + self.mle




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, unicode_literals, print_function

import re
import os
import tempfile
import contextlib
import multiprocessing

import six

from .. import environment
from ..console import log
from .. import util


WIN = (os.name == "nt")


# Conda (as of version 4.7.5) is not safe to run in parallel.
# See https://github.com/conda/conda/issues/8870
# Hence, serialize the calls to it.

util.new_multiprocessing_lock("conda_lock")

def _conda_lock():
    # function; for easier monkeypatching
    return util.get_multiprocessing_lock("conda_lock")


@contextlib.contextmanager
def _dummy_lock():
    yield


def _find_conda():
    """
    Find the conda executable robustly across conda versions.

    Returns
    -------
    conda : str
        Path to the conda executable.

    Raises
    ------
    IOError
        If the executable cannot be found in either the CONDA_EXE environment
        variable or in the PATH.

    Notes
    -----
    In POSIX platforms in conda >= 4.4, conda can be set up as a bash function
    rather than an executable. (This is to enable the syntax
    ``conda activate env-name``.) In this case, the environment variable
    ``CONDA_EXE`` contains the path to the conda executable. In other cases,
    we use standard search for the appropriate name in the PATH.

    See https://github.com/airspeed-velocity/asv/issues/645 for more details.
    """
    if 'CONDA_EXE' in os.environ:
        conda = os.environ['CONDA_EXE']
    else:
        conda = util.which('conda')
    return conda


class Conda(environment.Environment):
    """
    Manage an environment using conda.

    Dependencies are installed using ``conda``.  The benchmarked
    project is installed using ``pip`` (since ``conda`` doesn't have a
    method to install from an arbitrary ``setup.py``).
    """
    tool_name = "conda"
    _matches_cache = {}

    def __init__(self, conf, python, requirements, tagged_env_vars):
        """
        Parameters
        ----------
        conf : Config instance

        python : str
            Version of Python.  Must be of the form "MAJOR.MINOR".

        requirements : dict
            Dictionary mapping a PyPI package name to a version
            identifier string.
        """
        self._python = python
        self._requirements = requirements
        self._conda_channels = conf.conda_channels
        self._conda_environment_file = conf.conda_environment_file
        super(Conda, self).__init__(conf,
                                    python,
                                    requirements,
                                    tagged_env_vars)

    @classmethod
    def matches(cls, python):
        # Calling conda can take a long time, so remember the result
        if python not in cls._matches_cache:
            cls._matches_cache[python] = cls._matches(python)
        return cls._matches_cache[python]

    @classmethod
    def _matches(cls, python):
        if not re.match(r'^[0-9].*$', python):
            # The python name should be a version number
            return False

        try:
            conda = _find_conda()
        except IOError:
            return False
        else:
            # This directory never gets created, since we're just
            # doing a dry run below.  All it needs to be is something
            # that doesn't already exist.
            path = os.path.join(tempfile.gettempdir(), 'check')

            # Check that the version number is valid
            try:
                with _conda_lock():
                    util.check_call([
                        conda,
                        'create',
                        '--yes',
                        '-p',
                        path,
                        'python={0}'.format(python),
                        '--dry-run'], display_error=False, dots=False)
            except util.ProcessError:
                return False
            else:
                return True

    def _setup(self):
        log.info("Creating conda environment for {0}".format(self.name))

        conda_args, pip_args = self._get_requirements()
        env = dict(os.environ)
        env.update(self.build_env_vars)

        if not self._conda_environment_file:
            # The user-provided env file is assumed to set the python version
            conda_args = ['python={0}'.format(self._python), 'wheel', 'pip'] + conda_args

        # Create a temporary environment.yml file
        # and use that to generate the env for benchmarking.
        env_file = tempfile.NamedTemporaryFile(mode='w', delete=False, suffix=".yml")
        try:
            env_file.write('name: {0}\n'
                           'channels:\n'.format(self.name))
            env_file.writelines(('   - %s\n' % ch for ch in self._conda_channels))
            env_file.write('dependencies:\n')

            # categorize & write dependencies based on pip vs. conda
            env_file.writelines(('   - %s\n' % s for s in conda_args))
            if pip_args:
                # and now specify the packages that are to be installed in
                # the pip subsection
                env_file.write('   - pip:\n')
                env_file.writelines(('     - %s\n' % s for s in pip_args))

            env_file.close()

            try:
                env_file_name = self._conda_environment_file or env_file.name
                self._run_conda(['env', 'create', '-f', env_file_name,
                                 '-p', self._path, '--force'],
                                env=env)

                if self._conda_environment_file and (conda_args or pip_args):
                    # Add extra packages
                    env_file_name = env_file.name
                    self._run_conda(['env', 'update', '-f', env_file_name,
                                     '-p', self._path],
                                    env=env)
            except Exception:
                if env_file_name != env_file.name:
                    log.info("conda env create/update failed: in {} with file {}".format(self._path, env_file_name))
                elif os.path.isfile(env_file_name):
                    with open(env_file_name, 'r') as f:
                        text = f.read()
                    log.info("conda env create/update failed: in {} with:\n{}".format(self._path, text))
                raise
        finally:
            os.unlink(env_file.name)

    def _get_requirements(self):
        if self._requirements:
            # retrieve and return all conda / pip dependencies
            conda_args = []
            pip_args = []

            for key, val in six.iteritems(self._requirements):
                if key.startswith('pip+'):
                    if val:
                        pip_args.append("{0}=={1}".format(key[4:], val))
                    else:
                        pip_args.append(key[4:])
                else:
                    if val:
                        conda_args.append("{0}={1}".format(key, val))
                    else:
                        conda_args.append(key)

            return conda_args, pip_args
        else:
            return [], []

    def _run_conda(self, args, env=None):
        """
        Run conda command outside the environment.
        """
        try:
            conda = _find_conda()
        except IOError as e:
            raise util.UserError(str(e))

        with _conda_lock():
            return util.check_output([conda] + args, env=env)

    def run(self, args, **kwargs):
        log.debug("Running '{0}' in {1}".format(' '.join(args), self.name))
        return self.run_executable('python', args, **kwargs)

    def run_executable(self, executable, args, **kwargs):
        # Special-case running conda, for user-provided commands
        if executable == "conda":
            executable = _find_conda()
            lock = _conda_lock
        else:
            lock = _dummy_lock

        # Conda doesn't guarantee that user site directories are excluded
        kwargs["env"] = dict(kwargs.pop("env", os.environ),
                             PYTHONNOUSERSITE=str("True"))

        with lock():
            return super(Conda, self).run_executable(executable, args, **kwargs)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from ..benchmarks import Benchmarks
from . import Command
from ..repo import get_repo
from .setup import Setup
from ..console import log

from . import common_args
from .. import environment


class Check(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "check", help="Import and check benchmark suite, but do not run benchmarks",
            description="""
                This imports and checks basic validity of the benchmark suite, but
                does not run the benchmark target code""")

        common_args.add_environment(parser, default_same=False)
        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args, **kwargs):
        return cls.run(conf, env_spec=args.env_spec, **kwargs)

    @classmethod
    def run(cls, conf, env_spec=None):
        environments = list(environment.get_environments(conf, env_spec))

        if environment.is_existing_only(environments):
            # No repository required, so skip using it
            conf.dvcs = "none"

        repo = get_repo(conf)

        commit_hashes = []  # use fallback defaults
        benchmarks = Benchmarks.discover(conf, repo, environments, commit_hashes,
                                         check=True)
        return 0



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)


class OutputPublisher(object):
    """
    A base class for pages displaying output in the JS application
    """
    name = None
    button_label = None
    description = None
    order = float('inf')

    @classmethod
    def publish(cls, conf, repo, benchmarks, graphs, revisions):
        pass



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import six

from . import Command

from .. import machine


class Machine(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "machine", help="Define information about this machine",
            description="""
            Defines information about this machine.  If no arguments
            are provided, an interactive console session will be used
            to ask questions about the machine.
            """)

        defaults = machine.Machine.get_defaults()
        for name, description in machine.Machine.fields:
            parser.add_argument(
                '--' + name, default=defaults[name],
                help=description)

        parser.add_argument('--yes', default=False, action='store_true',
                            help="Accept all questions")

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args):
        return cls.run(**vars(args))

    @classmethod
    def run(cls, **kwargs):
        different = {}
        defaults = machine.Machine.get_defaults()
        for key, val in six.iteritems(defaults):
            if kwargs.get(key) != val:
                different[key] = kwargs.get(key)

        use_defaults = kwargs['yes']

        machine.Machine.load(
            force_interactive=(len(different) == 0),
            use_defaults=use_defaults, **different)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import sys

import six

from . import commands
from .config import Config
from .console import log
from .plugin_manager import plugin_manager
from . import util


def main():
    parser, subparsers = commands.make_argparser()

    args = parser.parse_args()

    if not hasattr(args, 'func'):
        parser.print_help()
        sys.exit(1)

    log.enable(args.verbose)

    args.config = os.path.abspath(args.config)

    # Use the path to the config file as the cwd for the remainder of
    # the run
    dirname = os.path.dirname(args.config)
    os.chdir(dirname)

    try:
        result = args.func(args)
    except util.UserError as e:
        log.error(six.text_type(e))
        sys.exit(1)
    finally:
        log.flush()

    if result is None:
        result = 0

    sys.exit(result)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, unicode_literals, print_function

import multiprocessing
import argparse

from .. import __version__
from .. import util


def add_global_arguments(parser, suppress_defaults=True):
    # Suppressing defaults is needed in order to allow global
    # arguments both before and after subcommand. Only the top-level
    # parser should have suppress_defaults=False

    if suppress_defaults:
        suppressor = dict(default=argparse.SUPPRESS)
    else:
        suppressor = dict()

    parser.add_argument(
        "--verbose", "-v", action="store_true",
        help="Increase verbosity",
        **suppressor)

    parser.add_argument(
        "--config",
        help="Benchmark configuration file",
        default=(argparse.SUPPRESS if suppress_defaults else 'asv.conf.json'))

    parser.add_argument(
        "--version", action="version", version="%(prog)s " + __version__,
        help="Print program version",
        **suppressor)


def add_compare(parser, only_changed_default=False, sort_default='name'):
    parser.add_argument(
        '--factor', "-f", type=float, default=1.1,
        help="""The factor above or below which a result is considered
        problematic.  For example, with a factor of 1.1 (the default
        value), if a benchmark gets 10%% slower or faster, it will
        be displayed in the results list.""")

    parser.add_argument(
        '--no-stats', action="store_false", dest="use_stats", default=True,
        help="""Do not use result statistics in comparisons, only `factor`
        and the median result.""")

    parser.add_argument(
        '--split', '-s', action='store_true',
        help="""Split the output into a table of benchmarks that have
        improved, stayed the same, and gotten worse.""")

    parser.add_argument(
        '--only-changed', action='store_true', default=only_changed_default,
        help="""Whether to show only changed results.""")

    parser.add_argument('--no-only-changed', dest='only_changed', action='store_false')

    parser.add_argument(
        '--sort', action='store', type=str, choices=('name', 'ratio'),
        default=sort_default, help="""Sort order""")


def add_show_stderr(parser):
    parser.add_argument(
        "--show-stderr", "-e", action="store_true",
        help="""Display the stderr output from the benchmarks.""")


class DictionaryArgAction(argparse.Action):
    """
    Parses multiple key=value assignments into a dictionary.
    """
    def __init__(self, option_strings, dest, converters=None, choices=None,
                 dict_dest=None, **kwargs):
        if converters is None:
            converters = {}
        self.converters = converters
        self.__choices = choices
        self.dict_dest = dict_dest
        super(DictionaryArgAction, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        # Parse and check value
        if self.dict_dest is None:
            try:
                key, value = values.split("=", 1)
            except ValueError:
                raise argparse.ArgumentError(self,
                                             "{!r} is not a key=value assignment".format(values))
        else:
            key = self.dict_dest
            value = values

        if self.__choices is not None and key not in self.__choices:
            raise argparse.ArgumentError(self,
                                         "{!r} cannot be set".format(key))

        dest_key = key
        conv = self.converters.get(key, None)
        if isinstance(conv, tuple):
            dest_key, conv = conv

        if conv is not None:
            try:
                value = conv(value)
            except ValueError as exc:
                raise argparse.ArgumentError(self,
                                             "{!r}: {}".format(key, exc))

        # Store value
        result = getattr(namespace, self.dest, None)
        if result is None:
            result = {}
        result[dest_key] = value
        setattr(namespace, self.dest, result)


def add_bench(parser):
    parser.add_argument(
        "--bench", "-b", type=str, action="append",
        help="""Regular expression(s) for benchmark to run.  When not
        provided, all benchmarks are run.""")

    def parse_repeat(value):
        try:
            return int(value)
        except ValueError:
            pass

        min_repeat, max_repeat, max_time = value.lstrip('(').rstrip(')').split(',')
        value = (int(min_repeat), int(max_repeat), float(max_time))
        return value

    def parse_affinity(value):
        if "," in value:
            value = value.split(",")
        else:
            value = [value]

        affinity_list = []
        for v in value:
            if "-" in v:
                a, b = v.split("-", 1)
                a = int(a)
                b = int(b)
                affinity_list.extend(range(a, b + 1))
            else:
                affinity_list.append(int(v))

        num_cpu = multiprocessing.cpu_count()
        for n in affinity_list:
            if not (0 <= n < num_cpu):
                raise ValueError("CPU {!r} not in range 0-{!r}".format(n, num_cpu-1))

        return affinity_list

    converters = {
        'timeout': float,
        'version': str,
        'warmup_time': float,
        'repeat': parse_repeat,
        'number': int,
        'rounds': int,
        'processes': ('rounds', int),  # backward compatibility
        'sample_time': float,
        'cpu_affinity': parse_affinity
    }

    parser.add_argument(
        "--attribute", "-a", action=DictionaryArgAction,
        choices=tuple(converters.keys()), converters=converters,
        help="""Override a benchmark attribute, e.g. `-a repeat=10`.""")

    parser.add_argument(
        "--cpu-affinity", action=DictionaryArgAction, dest="attribute",
        dict_dest="cpu_affinity",
        choices=tuple(converters.keys()), converters=converters,
        help=("Set CPU affinity for running the benchmark, in format: "
              "0 or 0,1,2 or 0-3. Default: not set"))


def add_machine(parser):
    parser.add_argument(
        "--machine", "-m", type=str, default=None,
        help="""Use the given name to retrieve machine information.
        If not provided, the hostname is used.  If no entry with that
        name is found, and there is only one entry in
        ~/.asv-machine.json, that one entry will be used.""")


class PythonArgAction(argparse.Action):
    """
    Backward compatibility --python XYZ argument,
    will be interpreted as --environment :XYZ
    """
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super(PythonArgAction, self).__init__(option_strings, dest, nargs=1, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        items = list(getattr(namespace, "env_spec", []))
        if values == "same":
            items.extend(["existing:same"])
        else:
            items.extend([":" + value for value in values])
        setattr(namespace, "env_spec", items)


def add_environment(parser, default_same=False):
    help = """Specify the environment and Python versions for running the
        benchmarks. String of the format 'environment_type:python_version',
        for example 'conda:2.7'. If the Python version is not specified,
        all those listed in the configuration file are run. The special
        environment type 'existing:/path/to/python' runs the benchmarks
        using the given Python interpreter; if the path is omitted,
        the Python running asv is used. For 'existing', the benchmarked
        project must be already installed, including all dependencies.
        """

    if default_same:
        help += "The default value is 'existing:same'"
    else:
        help += """By default, uses the values specified in the
            configuration file."""

    parser.add_argument(
        "-E", "--environment",
        dest="env_spec",
        action="append",
        default=[],
        help=help)

    # The --python argument exists for backward compatibility.  It
    # will just set the part after ':' in the environment spec.
    parser.add_argument(
        "--python", action=PythonArgAction, metavar="PYTHON",
        help="Same as --environment=:PYTHON")


def add_launch_method(parser):
    parser.add_argument(
        "--launch-method",
        dest="launch_method",
        action="store",
        choices=("auto", "spawn", "forkserver"),
        default="auto",
        help="How to launch benchmarks. Choices: auto, spawn, forkserver")


def add_parallel(parser):
    parser.add_argument(
        "--parallel", "-j", nargs='?', type=int, default=1, const=-1,
        help="""Build (but don't benchmark) in parallel.  The value is
        the number of CPUs to use, or if no number provided, use the
        number of cores on this machine.""")


def add_record_samples(parser, record_default=False):
    grp = parser.add_mutually_exclusive_group()
    grp.add_argument(
        "--record-samples", action="store_true", dest="record_samples",
        help=(argparse.SUPPRESS if record_default else
              """Store raw measurement samples, not only statistics"""),
        default=record_default)
    grp.add_argument(
        "--no-record-samples", action="store_false", dest="record_samples",
        help=(argparse.SUPPRESS if not record_default else
              """Do not store raw measurement samples, but only statistics"""),
        default=record_default)
    parser.add_argument(
        "--append-samples", action="store_true",
        help="""Combine new measurement samples with previous results,
        instead of discarding old results. Implies --record-samples.
        The previous run must also have been run with --record/append-samples.""")


def positive_int(string):
    """
    Parse a positive integer argument
    """
    try:
        value = int(string)
        if not value > 0:
            raise ValueError()
        return value
    except ValueError:
        raise argparse.ArgumentTypeError("%r is not a positive integer" % (string,))


def positive_int_or_inf(string):
    """
    Parse a positive integer argument
    """
    try:
        if string == 'all':
            return float("inf")
        value = int(string)
        if not value > 0:
            raise ValueError()
        return value
    except ValueError:
        raise argparse.ArgumentTypeError("%r is not a positive integer or 'all'" % (string,))


def time_period(string, base_period='d'):
    """
    Parse a time period argument with unit suffix
    """
    try:
        return util.parse_human_time(string, base_period)
    except ValueError as err:
        raise argparse.ArgumentTypeError(str(err))



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import json
import os
import re
import tempfile
import itertools
import datetime

import six

from .console import log
from . import util
from . import runner
from .repo import NoSuchNameError


class Benchmarks(dict):
    """
    Manages and runs the set of benchmarks in the project.
    """
    api_version = 2

    def __init__(self, conf, benchmarks, regex=None):
        """
        Initialize a list of benchmarks.

        Parameters
        ----------
        conf : Config object
            The project's configuration

        benchmarks : list
            Benchmarks as from Benchmarks._disc_benchmarks
            or loaded from a file.

        regex : str or list of str, optional
            `regex` is a list of regular expressions matching the
            benchmarks to run.  If none are provided, all benchmarks
            are run.
            For parameterized benchmarks, the regex match against
            `funcname(param0, param1, ...)` to include the parameter
            combination in regex filtering.
        """
        self._conf = conf
        self._benchmark_dir = conf.benchmark_dir

        if not regex:
            regex = []
        if isinstance(regex, six.string_types):
            regex = [regex]

        self._all_benchmarks = {}
        self._benchmark_selection = {}
        for benchmark in benchmarks:
            self._all_benchmarks[benchmark['name']] = benchmark
            if benchmark['params']:
                self._benchmark_selection[benchmark['name']] = []
                for idx, param_set in enumerate(
                        itertools.product(*benchmark['params'])):
                    name = '%s(%s)' % (
                        benchmark['name'],
                        ', '.join(param_set))
                    if not regex or any(re.search(reg, name) for reg in regex):
                        self[benchmark['name']] = benchmark
                        self._benchmark_selection[benchmark['name']].append(idx)
            else:
                self._benchmark_selection[benchmark['name']] = None
                if not regex or any(re.search(reg, benchmark['name']) for reg in regex):
                    self[benchmark['name']] = benchmark

    @property
    def benchmark_selection(self):
        """
        Active sets of parameterized benchmarks.
        """
        return self._benchmark_selection

    @property
    def benchmark_dir(self):
        """
        Benchmark directory.
        """
        return self._benchmark_dir

    def filter_out(self, skip):
        """
        Return a new Benchmarks object, with some benchmarks filtered out.
        """
        benchmarks = super(Benchmarks, self).__new__(self.__class__)
        benchmarks._conf = self._conf
        benchmarks._benchmark_dir = self._benchmark_dir
        benchmarks._all_benchmarks = self._all_benchmarks

        selected_idx = {}

        for name, benchmark in six.iteritems(self):
            if name not in skip:
                benchmarks[name] = benchmark
                if name in self._benchmark_selection:
                    selected_idx[name] = self._benchmark_selection[name]

        benchmarks._benchmark_selection = selected_idx

        return benchmarks

    @classmethod
    def discover(cls, conf, repo, environments, commit_hash, regex=None,
                 check=False):
        """
        Discover benchmarks in the given `benchmark_dir`.

        Parameters
        ----------
        conf : Config object
            The project's configuration

        repo : Repo object
            The project's repository

        environments : list of Environment
            List of environments available for benchmark discovery.

        commit_hash : list of str
            Commit hashes to use for benchmark discovery.

        regex : str or list of str, optional
            `regex` is a list of regular expressions matching the
            benchmarks to run.  If none are provided, all benchmarks
            are run.

        check : bool
            Run additional checks after discovery.

        """
        benchmarks = cls._disc_benchmarks(conf, repo, environments, commit_hash, check)
        return cls(conf, benchmarks, regex=regex)

    @classmethod
    def _disc_benchmarks(cls, conf, repo, environments, commit_hashes, check):
        """
        Discover all benchmarks in a directory tree.
        """
        root = conf.benchmark_dir

        cls.check_tree(root)

        if len(environments) == 0:
            raise util.UserError("No available environments")

        # Try several different commits:
        #
        # - First of commit_hashes provided
        # - Tips of branches from configuration file
        # - Rest of the commit_hashes
        #

        def iter_hashes():
            for h in commit_hashes[:1]:
                yield h
            for branch in conf.branches:
                try:
                    yield repo.get_hash_from_name(branch)
                except NoSuchNameError:
                    continue
            for h in commit_hashes[1:]:
                yield h

        def iter_unique(iter):
            seen = set()
            for item in iter:
                if item not in seen:
                    seen.add(item)
                    yield item

        try_hashes = iter_unique(iter_hashes())

        log.info("Discovering benchmarks")
        with log.indent():
            last_err = None
            for env, commit_hash in itertools.product(environments, try_hashes):
                env.create()

                if last_err is not None:
                    log.warning("Failed: trying different commit/environment")

                result_dir = tempfile.mkdtemp()
                try:
                    env.install_project(conf, repo, commit_hash)

                    env_vars = dict(os.environ)
                    env_vars.update(env.env_vars)

                    result_file = os.path.join(result_dir, 'result.json')
                    env.run(
                        [runner.BENCHMARK_RUN_SCRIPT, 'discover',
                         os.path.abspath(root),
                         os.path.abspath(result_file)],
                        cwd=result_dir,
                        env=env_vars,
                        dots=False)

                    try:
                        with open(result_file, 'r') as fp:
                            benchmarks = json.load(fp)
                    except (IOError, ValueError) as exc:
                        log.error("Invalid discovery output")
                        raise util.UserError()

                    break
                except (util.UserError, util.ProcessError) as err:
                    last_err = err
                    continue
                except KeyboardInterrupt:
                    raise util.UserError("Interrupted.")
                finally:
                    util.long_path_rmtree(result_dir)
            else:
                raise util.UserError("Failed to build the project and import the benchmark suite.")

        if check:
            log.info("Checking benchmarks")
            with log.indent():
                result_dir = tempfile.mkdtemp()
                try:
                    out, err, retcode = env.run(
                        [runner.BENCHMARK_RUN_SCRIPT, 'check',
                         os.path.abspath(root)],
                        cwd=result_dir,
                        dots=False,
                        env=env_vars,
                        valid_return_codes=None,
                        return_stderr=True,
                        redirect_stderr=True)
                finally:
                    util.long_path_rmtree(result_dir)

                out = out.strip()
                if retcode == 0:
                    if out:
                        log.info(out)
                    log.info("No problems found.")
                else:
                    if out:
                        log.error(out)
                    raise util.UserError("Benchmark suite check failed.")

        return benchmarks

    @classmethod
    def check_tree(cls, root, require_init_py=True):
        """
        Check the benchmark tree for files with the same name as
        directories.

        Also, ensure that the top-level directory has an __init__.py file.

        Raises
        ------
        UserError
            A .py file and directory with the same name (excluding the
            extension) were found.
        """
        if os.path.basename(root) == '__pycache__':
            return

        if not os.path.isfile(os.path.join(root, '__init__.py')):
            # Not a Python package directory
            if require_init_py:
                raise util.UserError(
                    "No __init__.py file in '{0}'".format(root))
            else:
                return

        # First, check for the case where a .py file and a directory
        # have the same name (without the extension).  This can't be
        # handled, so just raise an exception
        found = set()
        for filename in os.listdir(root):
            path = os.path.join(root, filename)
            if os.path.isfile(path):
                filename, ext = os.path.splitext(filename)
                if ext == '.py':
                    found.add(filename)

        for dirname in os.listdir(root):
            path = os.path.join(root, dirname)
            if os.path.isdir(path):
                if dirname in found:
                    raise util.UserError(
                        "Found a directory and python file with same name in "
                        "benchmark tree: '{0}'".format(path))
                cls.check_tree(path, require_init_py=False)

    @classmethod
    def get_benchmark_file_path(cls, results_dir):
        """
        Get the path to the benchmarks.json file in the results dir.
        """
        return os.path.join(results_dir, "benchmarks.json")

    def save(self):
        """
        Save the ``benchmarks.json`` file, which is a cached set of the
        metadata about the discovered benchmarks, in the results dir.
        """
        path = self.get_benchmark_file_path(self._conf.results_dir)
        util.write_json(path, self._all_benchmarks, self.api_version)

    @classmethod
    def load(cls, conf, regex=None):
        """
        Load the benchmark descriptions from the `benchmarks.json` file.
        If the file is not found, one of the given `environments` will
        be used to discover benchmarks.

        Parameters
        ----------
        conf : Config object
            The project's configuration
        regex : str or list of str, optional
            `regex` is a list of regular expressions matching the
            benchmarks to load. See __init__ docstring.

        Returns
        -------
        benchmarks : Benchmarks object
        """
        try:
            path = cls.get_benchmark_file_path(conf.results_dir)
            if not os.path.isfile(path):
                raise util.UserError("Benchmark list file {} missing!".format(path))
            d = util.load_json(path, api_version=cls.api_version)
            benchmarks = six.itervalues(d)
            return cls(conf, benchmarks, regex=regex)
        except util.UserError as err:
            if "asv update" in str(err):
                # Don't give conflicting instructions
                raise
            raise util.UserError("{}\nUse `asv run --bench just-discover` to "
                                 "regenerate benchmarks.json".format(str(err)))



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import math

from . import Command
from ..benchmarks import Benchmarks
from ..console import log
from ..machine import Machine
from ..repo import get_repo
from ..runner import run_benchmarks
from .. import util

from .setup import Setup

from . import common_args


def draw_graph(lo, mid, hi, total):
    nchars = 60
    scale = float(nchars) / total
    graph = ['-'] * nchars
    graph[int(lo * scale)] = '<'
    graph[int(hi * scale)] = '>'
    graph[int(mid * scale)] = 'O'
    return ''.join(graph)


class Find(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "find", help="Find commits that introduced large regressions",
            description="""Adaptively searches a range of commits for
            one that produces a large regression.  This only works well
            when the regression in the range is mostly monotonic.""")

        parser.add_argument(
            'range', type=str, metavar=('from..to',),
            help="""Range of commits to search.  For a git
            repository, this is passed as the first argument to ``git
            log``.  See 'specifying ranges' section of the
            `gitrevisions` manpage for more info.""")
        parser.add_argument(
            "bench", type=str, metavar=('benchmark_name',),
            help="""Name of benchmark to use in search.""")
        parser.add_argument(
            "--invert", "-i", action="store_true",
            help="""Search for a decrease in the benchmark value,
            rather than an increase.""")
        common_args.add_parallel(parser)
        common_args.add_show_stderr(parser)
        common_args.add_machine(parser)
        common_args.add_environment(parser)
        common_args.add_launch_method(parser)

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args, **kwargs):
        return cls.run(
            conf, args.range, args.bench,
            invert=args.invert, show_stderr=args.show_stderr,
            parallel=args.parallel,
            machine=args.machine, env_spec=args.env_spec,
            launch_method=args.launch_method, **kwargs
        )

    @classmethod
    def run(cls, conf, range_spec, bench, invert=False, show_stderr=False, parallel=1,
            machine=None, env_spec=None, _machine_file=None, launch_method=None):
        params = {}
        machine_params = Machine.load(
            machine_name=machine,
            _path=_machine_file,
            interactive=True)
        params.update(machine_params.__dict__)
        machine_params.save(conf.results_dir)

        repo = get_repo(conf)
        repo.pull()

        commit_hashes = repo.get_hashes_from_range(range_spec)[::-1]

        if len(commit_hashes) == 0:
            log.error("No commit hashes selected")
            return 1

        environments = Setup.run(conf=conf, env_spec=env_spec, parallel=parallel)
        if len(environments) == 0:
            log.error("No environments selected")
            return 1

        benchmarks = Benchmarks.discover(conf, repo, environments,
                                         commit_hashes, regex=bench)
        if len(benchmarks) == 0:
            log.error("'{0}' benchmark not found".format(bench))
            return 1
        elif len(benchmarks) > 1:
            exact_matches = benchmarks.filter_out([x for x in benchmarks if x != bench])
            if len(exact_matches) == 1:
                log.warning("'{0}' matches more than one benchmark, using exact match".format(bench))
                benchmarks = exact_matches
            else:
                log.error("'{0}' matches more than one benchmark".format(bench))
                return 1

        benchmark_name, = benchmarks.keys()
        benchmark_type = benchmarks[benchmark_name]["type"]

        steps = int(math.log(len(commit_hashes)) / math.log(2))

        log.info(
            "Running approximately {0} benchmarks within {1} commits".format(
                steps, len(commit_hashes)))

        env = environments[0]

        results = [None] * len(commit_hashes)

        def do_benchmark(i):
            if results[i] is not None:
                return results[i]

            commit_hash = commit_hashes[i]

            commit_name = repo.get_decorated_hash(commit_hash, 8)
            log.info(
                "For {0} commit {1}:".format(
                    conf.project, commit_name))

            env.install_project(conf, repo, commit_hash)

            res = run_benchmarks(benchmarks, env,
                                 show_stderr=show_stderr,
                                 launch_method=launch_method)

            result = res.get_result_value(benchmark_name,
                                          benchmarks[benchmark_name]['params'])

            results[i] = result

            # If we failed due to timeout in a timing benchmark, set
            # runtime as the timeout to prevent falling back to linear
            # search
            errcode = res.errcode[benchmark_name]
            if errcode == util.TIMEOUT_RETCODE and benchmark_type == "time":
                timeout_limit = benchmarks[benchmark_name]['timeout']
                results[i] = [r if r is not None else timeout_limit
                              for r in results[i]]

            return results[i]

        def non_null_results(*results):
            """
            Whether some value is non-null in all result sets
            """
            for values in zip(*results):
                if all(x is not None for x in values):
                    return True
            return False

        def difference_3way(a, b, c):
            """
            Return largest regression (a-b, b-c).
            """
            results_ab = [0]
            results_bc = [0]
            for va, vb, vc in zip(a, b, c):
                if va is not None and vb is not None and vc is not None:
                    denom = abs(va) + abs(vb) + abs(vc)
                    if denom == 0:
                        denom = 1.0
                    if invert:
                        denom*=-1.0

                    results_ab.append((va - vb) / denom)
                    results_bc.append((vb - vc) / denom)
            return max(results_ab), max(results_bc)

        def do_search(lo, hi):
            if hi - lo <= 1:
                return hi

            mid = int(math.floor((hi - lo) / 2) + lo)

            log.info(
                "Testing {0}".format(
                    draw_graph(lo, mid, hi, len(commit_hashes))))

            with log.indent():
                lo_result = None
                while lo_result is None:
                    lo_result = do_benchmark(lo)
                    if not non_null_results(lo_result):
                        lo_result = None
                        lo += 1
                        if lo >= mid:
                            raise util.UserError("Too many commits failed")

                mid_result = None
                while mid_result is None:
                    mid_result = do_benchmark(mid)
                    if not non_null_results(mid_result, lo_result):
                        mid_result = None
                        mid += 1
                        if mid >= hi:
                            raise util.UserError("Too many commits failed")

                hi_result = None
                while hi_result is None:
                    hi_result = do_benchmark(hi)
                    if not non_null_results(lo_result, mid_result, hi_result):
                        hi_result = None
                        hi -= 1
                        if hi <= mid:
                            raise util.UserError("Too many commits failed")

            diff_b, diff_a = difference_3way(hi_result, mid_result, lo_result)


            if diff_a >= diff_b:
                return do_search(lo, mid)
            else:
                return do_search(mid, hi)

        result = do_search(0, len(commit_hashes) - 1)

        commit_name = repo.get_decorated_hash(commit_hashes[result], 8)

        if invert:
            direction = "improvement"
        else:
            direction = "regression"

        log.info("Greatest {0} found: {1}".format(direction, commit_name))

        return 0



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Supports git repositories for the benchmarked project.
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import re
import shlex

from ..console import log
from ..repo import Repo, NoSuchNameError
from .. import util


class Git(Repo):
    dvcs = "git"
    _default_branch = "master"

    def __init__(self, url, mirror_path):
        self._git = util.which("git")
        self._path = os.path.abspath(mirror_path)
        self._pulled = False

        if self.is_local_repo(url):
            # Local repository, no need for mirror
            self._path = os.path.abspath(url)
            self._pulled = True
        elif not self.is_local_repo(self._path):
            if os.path.exists(self._path):
                self._raise_bad_mirror_error(self._path)

            # Clone is missing
            log.info("Cloning project")
            self._run_git(['clone', '--mirror', url, self._path],
                           cwd=None)

    @classmethod
    def is_local_repo(cls, path):
        return os.path.isdir(path) and (
            os.path.exists(os.path.join(path, '.git')) or
            os.path.isdir(os.path.join(path, 'objects')))

    @classmethod
    def url_match(cls, url):
        regexes = [
            r'^https?://.*?\.git$',
            r'^git@.*?\.git$']

        for regex in regexes:
            if re.match(regex, url):
                return True

        # Check for a local path
        if cls.is_local_repo(url):
            return True

        return False

    def _run_git(self, args, cwd=True, **kwargs):
        if cwd is True:
            cwd = self._path
        kwargs['cwd'] = cwd
        env = dict(kwargs.pop('env', os.environ))
        if cwd is not None:
            prev = env.get('GIT_CEILING_DIRECTORIES')
            env['GIT_CEILING_DIRECTORIES'] = os.pathsep.join(
                [os.path.join(os.path.abspath(cwd), os.pardir)]
                + ([prev] if prev is not None else []))
        return util.check_output([self._git] + args, env=env, **kwargs)

    def get_new_range_spec(self, latest_result, branch=None):
        return '{0}..{1}'.format(latest_result, self.get_branch_name(branch))

    def get_range_spec(self, commit_a, commit_b):
        return '{0}..{1}'.format(commit_a, commit_b)

    def pull(self):
        # We assume the remote isn't updated during the run of asv
        # itself.
        if self._pulled:
            return

        log.info("Fetching recent changes")
        self._run_git(['fetch', 'origin'])
        self._pulled = True

    def checkout(self, path, commit_hash):
        def checkout_existing(display_error):
            # Deinit fails if no submodules, so ignore its failure
            self._run_git(['submodule', 'deinit', '-f', '.'],
                          cwd=path, display_error=False, valid_return_codes=None)
            self._run_git(['checkout', '-f', commit_hash],
                          cwd=path, display_error=display_error)
            self._run_git(['clean', '-fdx'],
                          cwd=path, display_error=display_error)
            self._run_git(['submodule', 'update', '--init', '--recursive'],
                          cwd=path, display_error=display_error)

        if os.path.isdir(path):
            try:
                checkout_existing(display_error=False)
            except util.ProcessError:
                # Remove and try to re-clone
                util.long_path_rmtree(path)

        if not os.path.isdir(path):
            self._run_git(['clone', '--shared', '--recursive', self._path, path],
                          cwd=None)
            checkout_existing(display_error=True)

    def get_date(self, hash):
        return int(self._run_git(
            ['rev-list', '-n', '1', '--format=%at', hash],
            valid_return_codes=(0, 1), dots=False).strip().split()[-1]) * 1000

    def get_hashes_from_range(self, range_spec):
        args = ['rev-list', '--first-parent']
        if range_spec != "":
            args += shlex.split(range_spec)
        output = self._run_git(args, valid_return_codes=(0, 1), dots=False)
        return output.strip().split()

    def get_hash_from_name(self, name):
        if name is None:
            name = self.get_branch_name()

        # In case of annotated tags, return the hash for the commit
        lookup_name = name + '^{commit}'

        try:
            return self._run_git(['rev-parse', lookup_name],
                                 display_error=False,
                                 dots=False).strip().split()[0]
        except util.ProcessError as err:
            if err.stdout.strip() == lookup_name:
                # Name does not exist
                raise NoSuchNameError(name)
            raise

    def get_hash_from_parent(self, name):
        return self.get_hash_from_name(name + '^')

    def get_name_from_hash(self, commit):
        try:
            name = self._run_git(["name-rev", "--name-only",
                                  "--exclude=remotes/*",
                                  "--no-undefined", commit],
                                 display_error=False).strip()
            if not name:
                return None
        except util.ProcessError as err:
            # Failed to obtain.
            return None

        # Return tags without prefix
        for prefix in ['tags/']:
            if name.startswith(prefix):
                return name[len(prefix):]

        return name

    def get_tags(self):
        tags = {}
        for tag in self._run_git(["tag", "-l"]).splitlines():
            tags[tag] = self._run_git(["rev-list", "-n", "1", tag]).strip()
        return tags

    def get_date_from_name(self, name):
        return self.get_date(name + "^{commit}")

    def get_branch_commits(self, branch):
        return self.get_hashes_from_range(self.get_branch_name(branch))

    def get_revisions(self, commits):
        revisions = {}
        for i, commit in enumerate(self._run_git([
            "rev-list", "--all", "--date-order", "--reverse",
        ]).splitlines()):
            if commit in commits:
                revisions[commit] = i
        return revisions



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import itertools

from . import Command
from ..benchmarks import Benchmarks
from ..machine import iter_machine_files
from ..results import iter_results_for_machine_and_hash
from ..repo import get_repo, NoSuchNameError
from ..util import human_value, load_json
from ..console import log, color_print
from ..environment import get_environments
from .. import util
from .. import statistics

from . import common_args


def mean(values):
    if all([value is None for value in values]):
        return None
    else:
        values = [value for value in values if value is not None]
        return sum(values) / float(len(values))


def unroll_result(benchmark_name, params, *values):
    """
    Iterate through parameterized result values

    Yields
    ------
    name
        Strings of the form "benchmark_name(value1, value2)" with
        parameter values substituted in. For non-parameterized
        results, simply the benchmark name.
    value
        Benchmark timing or other scalar value.

    """
    num_comb = 1
    for p in params:
        num_comb *= len(p)

    values = list(values)
    for j in range(len(values)):
        if values[j] is None:
            values[j] = [None] * num_comb

    for params, value in zip(itertools.product(*params), zip(*values)):
        if params == ():
            name = benchmark_name
        else:
            name = "%s(%s)" % (benchmark_name, ", ".join(params))
        yield (name,) + value


def _isna(value):
    # None (failed) or NaN (skipped)
    return value is None or value != value


def _is_result_better(a, b, a_ss, b_ss, factor, use_stats=True):
    """
    Check if result 'a' is better than 'b' by the given factor,
    possibly taking confidence intervals into account.

    """

    if use_stats and a_ss and b_ss and a_ss[0] and b_ss[0] and (
            a_ss[0].get('repeat', 0) != 1 and b_ss[0].get('repeat', 0) != 1):
        # Return False if estimates don't differ.
        #
        # Special-case the situation with only one sample, in which
        # case we do the comparison only based on `factor` as there's
        # not enough data to do statistics.
        if not statistics.is_different(a_ss[1], b_ss[1],
                                       a_ss[0], b_ss[0]):
            return False

    return a < b / factor


class Compare(Command):

    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "compare",
            help="""Compare the benchmark results between two revisions
                    (averaged over configurations)""",
            description="Compare two sets of results")

        parser.add_argument(
            'revision1',
            help="""The reference revision.""")

        parser.add_argument(
            'revision2',
            help="""The revision being compared.""")

        common_args.add_compare(parser, sort_default='name', only_changed_default=False)

        parser.add_argument(
            '--machine', '-m', type=str, default=None,
            help="""The machine to compare the revisions for.""")

        common_args.add_environment(parser)

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args):
        return cls.run(conf=conf,
                       hash_1=args.revision1,
                       hash_2=args.revision2,
                       factor=args.factor, split=args.split,
                       only_changed=args.only_changed, sort=args.sort,
                       machine=args.machine,
                       env_spec=args.env_spec,
                       use_stats=args.use_stats)

    @classmethod
    def run(cls, conf, hash_1, hash_2, factor=None, split=False, only_changed=False,
            sort='name', machine=None, env_spec=None, use_stats=True):

        repo = get_repo(conf)
        try:
            hash_1 = repo.get_hash_from_name(hash_1)
        except NoSuchNameError:
            pass

        try:
            hash_2 = repo.get_hash_from_name(hash_2)
        except NoSuchNameError:
            pass

        if env_spec:
            env_names = ([env.name for env in get_environments(conf, env_spec, verbose=False)]
                         + list(env_spec))
        else:
            env_names = None

        machines = []
        for path in iter_machine_files(conf.results_dir):
            d = load_json(path)
            machines.append(d['machine'])

        if len(machines) == 0:
            raise util.UserError("No results found")
        elif machine is None:
            if len(machines) > 1:
                raise util.UserError(
                    "Results available for several machines: {0} - "
                    "specify which one to use with the --machine option".format(
                        '/'.join(machines)))
            else:
                machine = machines[0]
        elif not machine in machines:
            raise util.UserError(
                "Results for machine '{0} not found".format(machine))

        commit_names = {hash_1: repo.get_name_from_hash(hash_1),
                        hash_2: repo.get_name_from_hash(hash_2)}

        cls.print_table(conf, hash_1, hash_2, factor=factor, split=split, use_stats=use_stats,
                        only_changed=only_changed, sort=sort,
                        machine=machine, env_names=env_names, commit_names=commit_names)

    @classmethod
    def print_table(cls, conf, hash_1, hash_2, factor, split,
                    resultset_1=None, resultset_2=None, machine=None,
                    only_changed=False, sort='name', use_stats=True, env_names=None,
                    commit_names=None):
        results_1 = {}
        results_2 = {}
        ss_1 = {}
        ss_2 = {}
        versions_1 = {}
        versions_2 = {}
        units = {}

        benchmarks = Benchmarks.load(conf)

        if commit_names is None:
            commit_names = {}

        def results_default_iter(commit_hash):
            for result in iter_results_for_machine_and_hash(
                    conf.results_dir, machine, commit_hash):
                if env_names is not None and result.env_name not in env_names:
                    continue
                for key in result.get_all_result_keys():
                    params = result.get_result_params(key)
                    result_value = result.get_result_value(key, params)
                    result_stats = result.get_result_stats(key, params)
                    result_samples = result.get_result_samples(key, params)
                    result_version = result.benchmark_version.get(key)
                    yield (key, params, result_value, result_stats, result_samples,
                           result_version, result.params['machine'], result.env_name)

        if resultset_1 is None:
            resultset_1 = results_default_iter(hash_1)

        if resultset_2 is None:
            resultset_2 = results_default_iter(hash_2)

        machine_env_names = set()

        for key, params, value, stats, samples, version, machine, env_name in resultset_1:
            machine_env_name = "{}/{}".format(machine, env_name)
            machine_env_names.add(machine_env_name)
            for name, value, stats, samples in unroll_result(key, params, value, stats, samples):
                units[(name, machine_env_name)] = benchmarks.get(key, {}).get('unit')
                results_1[(name, machine_env_name)] = value
                ss_1[(name, machine_env_name)] = (stats, samples)
                versions_1[(name, machine_env_name)] = version

        for key, params, value, stats, samples, version, machine, env_name in resultset_2:
            machine_env_name = "{}/{}".format(machine, env_name)
            machine_env_names.add(machine_env_name)
            for name, value, stats, samples in unroll_result(key, params, value, stats, samples):
                units[(name, machine_env_name)] = benchmarks.get(key, {}).get('unit')
                results_2[(name, machine_env_name)] = value
                ss_2[(name, machine_env_name)] = (stats, samples)
                versions_2[(name, machine_env_name)] = version

        if len(results_1) == 0:
            raise util.UserError(
                "Did not find results for commit {0}".format(hash_1))

        if len(results_2) == 0:
            raise util.UserError(
                "Did not find results for commit {0}".format(hash_2))

        benchmarks_1 = set(results_1.keys())
        benchmarks_2 = set(results_2.keys())

        joint_benchmarks = sorted(list(benchmarks_1 | benchmarks_2))

        bench = {}

        if split:
            bench['green'] = []
            bench['red'] = []
            bench['lightgrey'] = []
            bench['default'] = []
        else:
            bench['all'] = []

        worsened = False
        improved = False

        for benchmark in joint_benchmarks:
            if benchmark in results_1:
                time_1 = results_1[benchmark]
            else:
                time_1 = float("nan")

            if benchmark in results_2:
                time_2 = results_2[benchmark]
            else:
                time_2 = float("nan")

            if benchmark in ss_1 and ss_1[benchmark][0]:
                err_1 = statistics.get_err(time_1, ss_1[benchmark][0])
            else:
                err_1 = None

            if benchmark in ss_2 and ss_2[benchmark][0]:
                err_2 = statistics.get_err(time_2, ss_2[benchmark][0])
            else:
                err_2 = None

            version_1 = versions_1.get(benchmark)
            version_2 = versions_2.get(benchmark)

            if _isna(time_1) or _isna(time_2):
                ratio = 'n/a'
                ratio_num = 1e9
            else:
                try:
                    ratio_num = time_2 / time_1
                    ratio = "{0:6.2f}".format(ratio_num)
                except ZeroDivisionError:
                    ratio_num = 1e9
                    ratio = "n/a"

            if (version_1 is not None and version_2 is not None and
                    version_1 != version_2):
                # not comparable
                color = 'lightgrey'
                mark = 'x'
            elif time_1 is not None and time_2 is None:
                # introduced a failure
                color = 'red'
                mark = '!'
                worsened = True
            elif time_1 is None and time_2 is not None:
                # fixed a failure
                color = 'green'
                mark = ' '
                improved = True
            elif time_1 is None and time_2 is None:
                # both failed
                color = 'default'
                mark = ' '
            elif _isna(time_1) or _isna(time_2):
                # either one was skipped
                color = 'default'
                mark = ' '
            elif _is_result_better(time_2, time_1,
                                   ss_2.get(benchmark), ss_1.get(benchmark),
                                   factor, use_stats=use_stats):
                color = 'green'
                mark = '-'
                improved = True
            elif _is_result_better(time_1, time_2,
                                   ss_1.get(benchmark), ss_2.get(benchmark),
                                   factor, use_stats=use_stats):
                color = 'red'
                mark = '+'
                worsened = True
            else:
                color = 'default'
                mark = ' '

                # Mark statistically insignificant results
                if (_is_result_better(time_1, time_2, None, None, factor) or
                        _is_result_better(time_2, time_1, None, None, factor)):
                    ratio = "~" + ratio.strip()

            if only_changed and mark in (' ', 'x'):
                continue

            unit = units[benchmark]

            details = "{0:1s} {1:>15s}  {2:>15s} {3:>8s}  ".format(
                mark,
                human_value(time_1, unit, err=err_1),
                human_value(time_2, unit, err=err_2),
                ratio)

            if split:
                bench[color].append((color, details, benchmark, ratio_num))
            else:
                bench['all'].append((color, details, benchmark, ratio_num))

        if split:
            keys = ['green', 'default', 'red', 'lightgrey']
        else:
            keys = ['all']

        titles = {}
        titles['green'] = "Benchmarks that have improved:"
        titles['default'] = "Benchmarks that have stayed the same:"
        titles['red'] = "Benchmarks that have got worse:"
        titles['lightgrey'] = "Benchmarks that are not comparable:"
        titles['all'] = "All benchmarks:"

        log.flush()

        for key in keys:

            if len(bench[key]) == 0:
                continue

            if not only_changed:
                color_print("")
                color_print(titles[key])
                color_print("")
            color_print("       before           after         ratio")
            color_print("     [{0:8s}]       [{1:8s}]".format(hash_1[:8], hash_2[:8]))

            name_1 = commit_names.get(hash_1)
            if name_1:
                name_1 = '<{0}>'.format(name_1)
            else:
                name_1 = ''

            name_2 = commit_names.get(hash_2)
            if name_2:
                name_2 = '<{0}>'.format(name_2)
            else:
                name_2 = ''

            if name_1 or name_2:
                color_print("     {0:10s}       {1:10s}".format(name_1, name_2))

            if sort == 'ratio':
                bench[key].sort(key=lambda v: v[3], reverse=True)
            elif sort == 'name':
                bench[key].sort(key=lambda v: v[2])
            else:
                raise ValueError("Unknown 'sort'")

            for color, details, benchmark, ratio in bench[key]:
                if len(machine_env_names) > 1:
                    benchmark_name = "{} [{}]".format(*benchmark)
                else:
                    benchmark_name = benchmark[0]

                color_print(details, color, end='')
                color_print(benchmark_name)

        return worsened, improved



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os

from ..commands import Command
from ..commands.publish import Publish
from ..console import log
from .. import util


class GithubPages(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "gh-pages", help="Publish the results to Github pages",
            description="""
            Publish the results to github pages

            Updates the 'gh-pages' branch in the current repository,
            and pushes it to 'origin'.
            """)
        parser.add_argument(
            "--no-push", action="store_true",
            help="Update local gh-pages branch but don't push")
        parser.add_argument(
            "--rewrite", action="store_true",
            help=("Rewrite gh-pages branch to contain only a single commit, "
                  "instead of adding a new commit"))

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args):
        return cls.run(conf=conf, no_push=args.no_push, rewrite=args.rewrite)

    @classmethod
    def run(cls, conf, no_push, rewrite):
        git = util.which('git')

        # Publish
        Publish.run(conf)

        cwd = os.path.abspath(os.getcwd())

        log.info("Updating gh-pages branch")
        try:
            # Create new repo for the html data
            util.check_call([git, 'init'], cwd=conf.html_dir)
            util.check_call([git, 'checkout', '-b', 'gh-pages'], cwd=conf.html_dir)

            if not rewrite:
                # Fetch the old branch
                _, _, retcode = util.check_output([git, 'fetch', cwd, 'gh-pages'],
                                                  cwd=conf.html_dir,
                                                  return_stderr=True,
                                                  valid_return_codes=None)
                if retcode == 0:
                    util.check_call([git, 'reset', '--soft', 'FETCH_HEAD'], cwd=conf.html_dir)

            # We need to tell github this is not a Jekyll document
            with open(os.path.join(conf.html_dir, '.nojekyll'), 'wb') as fd:
                fd.write(b'\n')

            # Add all files
            util.check_call([git, 'add', '--all', '.'], cwd=conf.html_dir)

            # Commit (if needed)
            _, _, retcode = util.check_output([git, "diff-index", "--quiet", "HEAD", "--"],
                                              cwd=conf.html_dir,
                                              return_stderr=True,
                                              valid_return_codes=None)
            if retcode != 0:
                util.check_call([git, 'commit', '-m', 'Generated from sources'], cwd=conf.html_dir)

            # Fetch branch here
            if rewrite:
                util.check_call([git, 'fetch', os.path.abspath(conf.html_dir)])
                util.check_call([git, 'branch', '-f', 'gh-pages', 'FETCH_HEAD'])
            else:
                util.check_call([git, 'fetch', os.path.abspath(conf.html_dir),
                                 'gh-pages:gh-pages'])
        finally:
            # Cleanup the child repo under html
            util.long_path_rmtree(os.path.join(conf.html_dir, '.git'))

        # Push branch
        if not no_push:
            if rewrite:
                log.info("Force-pushing gh-pages branch")
                util.check_call([git, 'push', '-f', 'origin', 'gh-pages'])
            else:
                log.info("Pushing gh-pages branch")
                util.check_call([git, 'push', 'origin', 'gh-pages'])



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.


class TimeSuite:
    """
    An example benchmark that times the performance of various kinds
    of iterating over dictionaries in Python.
    """
    def setup(self):
        self.d = {}
        for x in range(500):
            self.d[x] = None

    def time_keys(self):
        for key in self.d.keys():
            pass

    def time_values(self):
        for value in self.d.values():
            pass

    def time_range(self):
        d = self.d
        for key in range(500):
            x = d[key]


class MemSuite:
    def mem_list(self):
        return [0] * 256



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import re

from . import Command
from ..config import Config
from ..machine import Machine, MachineCollection
from ..results import Results, get_filename
from ..benchmarks import Benchmarks
from ..console import log
from .. import util
from .run import Run


class Update(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "update", help="Update the results and config files "
            "to the current version",
            description="Update the results and config files "
            "to the current version")

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args, _machine_file=None):
        return cls.run(conf, _machine_file=_machine_file)

    @classmethod
    def run(cls, conf, _machine_file=None):
        MachineCollection.update(_path=_machine_file)

        log.info("Updating results data...")

        for root, dirs, files in os.walk(conf.results_dir):
            for filename in files:
                path = os.path.join(root, filename)
                if filename == 'machine.json':
                    Machine.update(path)
                elif filename == "benchmarks.json":
                    pass
                elif filename.endswith('.json'):
                    try:
                        Results.update(path)
                    except util.UserError as err:
                        # Conversion failed: just skip the file
                        log.warning("{}: {}".format(path, err))
                        continue

                    # Rename files if necessary
                    m = re.match(r'^([0-9a-f]+)-(.*)\.json$', os.path.basename(path), re.I)
                    if m:
                        new_path = get_filename(root, m.group(1), m.group(2))
                        if new_path != path:
                            try:
                                if os.path.exists(new_path):
                                    raise OSError()
                                os.rename(path, new_path)
                            except OSError:
                                log.warning("{}: should be renamed to {}".format(path, new_path))
                    else:
                        log.warning("{}: unrecognized file name".format(path))

        # Check benchmarks.json
        log.info("Updating benchmarks.json...")
        ok = False
        try:
            Benchmarks.load(conf)
            ok = True
        except util.UserError:
            pass

        if not ok:
            # Regenerating the file is needed
            with log.indent():
                Run.run(conf, bench=['just-discover'])
